import numpy as np
import copy

from mbo_utilities.roi import iter_rois
from mbo_utilities.array_types import DemixingResultsArray, MBOTiffArray, MboRawArray
import fastplotlib as fpl

from mbo_utilities.pipelines import MBO_DEVICE, HAS_MASKNMF

if HAS_MASKNMF:
    from masknmf.visualization.interactive_guis import make_demixing_video
else:
    make_demixing_video = None


def mbo_tiff_display(array: MBOTiffArray, **kwargs):
    return fpl.ImageWidget(
        data=array,
        histogram_widget=True,
        figure_kwargs={
            "size": (800, 1000),
        },  # "canvas": canvas},
        graphic_kwargs={"vmin": array.min(), "vmax": array.max()},
        window_funcs={"t": (np.mean, 0)},
    )


def mbo_raw_display(array: MboRawArray, **kwargs):
    arrays = []
    names = []
    # if roi is None, use a single array.roi = None
    # if roi is 0, get a list of all ROIs by deeepcopying the array and setting each roi
    for roi in iter_rois(array):
        arr = copy.copy(array)
        arr.roi = roi
        arrays.append(arr)
        names.append(f"ROI {roi}" if roi else "Full Image")

    return fpl.ImageWidget(
        data=arrays,
        names=names,
        histogram_widget=True,
        figure_kwargs={
            "size": (800, 1000),
        },  # "canvas": canvas},
        graphic_kwargs={"vmin": array.min(), "vmax": array.max()},
        window_funcs={"t": (np.mean, 0)},
    )

def imshow_lazy_array(array, **kwargs):
    if hasattr(array, "imshow"):
        return array.imshow(**kwargs)
    if isinstance(array, MBOTiffArray):
        return mbo_tiff_display(array, **kwargs)
    elif isinstance(array, MboRawArray):
        return mbo_raw_display(array, **kwargs)
    raise ValueError("No supported lazy array type found for display.")



import functools
import warnings
from typing import Any

import numpy as np

import shutil
from pathlib import Path
from tifffile import TiffWriter
import h5py

from . import log
from .file_io import write_ops
from ._parsing import _make_json_serializable
from .util import is_running_jupyter

try:
    from suite2p.io import BinaryFile

    HAS_SUITE2P = True
except ImportError:
    HAS_SUITE2P = False
    BinaryFile = None

if is_running_jupyter():
    from tqdm.notebook import tqdm
else:
    from tqdm.auto import tqdm

logger = log.get("writers")

warnings.filterwarnings("ignore")

ARRAY_METADATA = ["dtype", "shape", "nbytes", "size"]
CHUNKS = {0: "auto", 1: -1, 2: -1}


def _close_bin_writers():
    if hasattr(_write_bin, "_writers"):
        for bf in _write_bin._writers.values():
            bf.close()
        _write_bin._writers.clear()
        _write_bin._offsets.clear()


def _close_tiff_writers():
    if hasattr(_write_tiff, "_writers"):
        for writer in _write_tiff._writers.values():
            writer.close()
        _write_tiff._writers.clear()


def _write_plane(
    data: np.ndarray | Any,
    filename: Path,
    *,
    overwrite=False,
    metadata=None,
    target_chunk_mb=20,
    progress_callback=None,
    debug=False,
    dshape=None,
    plane_index=None,
):
    if dshape is None:
        dshape = data.shape

    metadata = metadata or {}
    metadata["shape"] = dshape

    if plane_index:
        assert type(plane_index) is int, "plane_index must be an integer"
        metadata["plane"] = plane_index + 1

    fname = filename
    writer = _get_file_writer(fname.suffix, overwrite=overwrite)

    chunk_size = target_chunk_mb * 1024 * 1024
    nbytes = np.prod(dshape) * np.dtype(data.dtype).itemsize
    nchunks = max(1, int(np.ceil(nbytes / chunk_size)))

    # this works for txy and tzxy
    ntime = data.shape[0]
    base = ntime // nchunks
    extra = ntime % nchunks

    if not debug:
        pbar = tqdm(total=nchunks, desc=f"Saving {fname.name}")
    else:
        pbar = None

    start = 0
    for i in range(nchunks):
        end = start + base + (1 if i < extra else 0)
        chunk = data[start:end, plane_index, :, :] if plane_index is not None else data[start:end, :, :]
        writer(fname, chunk, metadata=metadata)
        if pbar:
            pbar.update(1)
        if progress_callback:
            progress_callback(pbar.n / pbar.total, current_plane=plane_index)
        start = end

    if pbar:
        pbar.close()

    if fname.suffix in [".tiff", ".tif"]:
        _close_tiff_writers()
    elif fname.suffix in [".bin"]:
        _close_bin_writers()

def _get_file_writer(ext, overwrite):
    if ext.startswith("."):
        ext = ext.lstrip(".")
    if ext in ["tif", "tiff"]:
        return functools.partial(
            _write_tiff,
            overwrite=overwrite,
        )
    elif ext in ["h5", "hdf5"]:
        return functools.partial(
            _write_h5,
            overwrite=overwrite,
        )
    elif ext in ["zarr"]:
        return functools.partial(
            _write_zarr,
            overwrite=overwrite,
        )
    elif ext == "bin":
        if not HAS_SUITE2P:
            raise ValueError(
                "Suite2p needed to write binary files, please install it:\n"
                "pip install suite2p[io]"
            )
        return functools.partial(
            _write_bin,
            overwrite=overwrite,
        )
    else:
        raise ValueError(f"Unsupported file extension: {ext}")


def _write_bin(path, data, *, overwrite: bool = False, metadata=None):
    if not hasattr(_write_bin, "_writers"):
        _write_bin._writers, _write_bin._offsets = {}, {}

    fname = Path(path)
    fname.parent.mkdir(exist_ok=True)

    key = str(fname)
    first_write = False
    if key not in _write_bin._writers:
        if overwrite and fname.exists():
            fname.unlink()

        Ly, Lx = data.shape[1], data.shape[2]
        nframes = metadata.get("nframes", None)
        if nframes is None:
            nframes = metadata.get("num_frames", None)
        if nframes is None:
            raise ValueError("Metadata must contain 'nframes' or 'num_frames'.")

        _write_bin._writers[key] = BinaryFile(
            Ly, Lx, key, n_frames=metadata["num_frames"], dtype=np.int16
        )
        _write_bin._offsets[key] = 0
        first_write = True

    bf = _write_bin._writers[key]
    off = _write_bin._offsets[key]
    bf[off : off + data.shape[0]] = data
    bf.file.flush()
    _write_bin._offsets[key] = off + data.shape[0]

    if first_write:
        raw_filename = fname  # points to data_raw.bin
        write_ops(metadata, raw_filename)

    logger.debug(f"Wrote {data.shape[0]} frames to {fname}.")


def _write_h5(path, data, *, overwrite=True, metadata=None):
    filename = Path(path).with_suffix(".h5")

    if not hasattr(_write_h5, "_initialized"):
        _write_h5._initialized = {}
        _write_h5._offsets = {}

    if filename not in _write_h5._initialized:
        nframes = metadata.get("num_frames", None)
        if nframes is None:
            raise ValueError("Metadata must contain 'nframes' or 'nun_frames'.")
        h, w = data.shape[-2:]
        with h5py.File(filename, "w" if overwrite else "a") as f:
            f.create_dataset(
                "mov",
                shape=(nframes, h, w),
                maxshape=(None, h, w),
                chunks=(1, h, w),
                dtype=data.dtype,
                compression=None,
            )
            if metadata:
                for k, v in metadata.items():
                    f.attrs[k] = v if np.isscalar(v) else str(v)

        _write_h5._initialized[filename] = True
        _write_h5._offsets[filename] = 0

    offset = _write_h5._offsets[filename]
    with h5py.File(filename, "a") as f:
        f["mov"][offset : offset + data.shape[0]] = data

    _write_h5._offsets[filename] = offset + data.shape[0]


def _write_tiff(
    path, data, overwrite=True, metadata=None,
):
    filename = Path(path).with_suffix(".tif")

    if not hasattr(_write_tiff, "_writers"):
        _write_tiff._writers = {}
    if not hasattr(_write_tiff, "_first_write"):
        _write_tiff._first_write = {}

    if filename not in _write_tiff._writers:
        if filename.exists() and overwrite:
            filename.unlink()
        _write_tiff._writers[filename] = TiffWriter(filename, bigtiff=True)
        _write_tiff._first_write[filename] = True

    writer = _write_tiff._writers[filename]
    is_first = _write_tiff._first_write.get(filename, True)

    for frame in data:
        writer.write(
            frame,
            contiguous=True,
            photometric="minisblack",
            metadata=_make_json_serializable(metadata) if is_first else None,
        )
        _write_tiff._first_write[filename] = False


def _write_zarr(path, data, *, overwrite=True, metadata=None):
    filename = Path(path).with_suffix(".zarr")

    if not hasattr(_write_zarr, "_arrays"):
        _write_zarr._arrays = {}
        _write_zarr._offsets = {}

    if filename not in _write_zarr._arrays:
        if filename.exists() and overwrite:
            shutil.rmtree(filename)

        import zarr

        nframes = metadata["num_frames"]
        h, w = data.shape[-2:]
        z = zarr.open(
            store=str(filename),
            mode="w",
            shape=(nframes, h, w),
            chunks=(1, h, w),
            dtype=data.dtype,
        )
        if metadata:
            for k, v in metadata.items():
                try:
                    z.attrs[k] = v
                except TypeError:
                    z.attrs[k] = str(v)

        _write_zarr._arrays[filename] = z
        _write_zarr._offsets[filename] = 0

    z = _write_zarr._arrays[filename]
    offset = _write_zarr._offsets[filename]
    z[offset : offset + data.shape[0]] = data
    _write_zarr._offsets[filename] = offset + data.shape[0]




from pathlib import Path

from .file_io import (
    get_files,
    npy_to_dask,
    expand_paths,
    get_mbo_dirs,
    load_ops,
    write_ops,
    get_plane_from_filename,
)
from ._parsing import _normalize_file_url
from .plot_util import save_png, save_mp4

from .metadata import is_raw_scanimage, get_metadata, params_from_metadata
from .util import (
    norm_minmax,
    smooth_data,
    is_running_jupyter,
    is_imgui_installed,
    subsample_array,
)
from .lazy_array import imread, imwrite


if is_imgui_installed():
    from .graphics import run_gui
else:
    raise ImportError(
        f"This should be installed with mbo_utilities. Please report this [here](https://github.com/MillerBrainObservatory/mbo_utilities/issues) or on slack."
    )

__version__ = (Path(__file__).parent / "VERSION").read_text().strip()


__all__ = [
    # file_io
    "imread",
    "imwrite",
    "run_gui",
    "get_mbo_dirs",
    "scanreader",
    "npy_to_dask",
    "get_files",
    "subsample_array",
    "load_ops",
    "write_ops",
    "get_plane_from_filename",
    # metadata
    "is_raw_scanimage",
    "get_metadata",
    "params_from_metadata",
    # util
    "expand_paths",
    "norm_minmax",
    "smooth_data",
    "is_running_jupyter",
    "is_imgui_installed",  # we may just enforce imgui?
    # assembly
    "save_mp4",
    "save_png",
]



import os
import functools

import pytest
from pathlib import Path

import numpy as np
from icecream import ic
from tifffile import imread as tfile_imread

import mbo_utilities as mbo
from mbo_utilities import get_mbo_dirs
from mbo_utilities._benchmark import _benchmark_indexing
from mbo_utilities.lazy_array import imread as mbo_imread, imwrite as mbo_imwrite

try:
    import dask.array as da
    import zarr
except ImportError:
    raise ImportError(
        "Dask and Zarr are required for this test. "
        "Please install them with `pip install dask zarr`."
    )

ic.enable()

DEFAULT_DATA_ROOT = get_mbo_dirs()["tests"]
DATA_ROOT = Path(os.getenv("MBO_TEST_DATA", DEFAULT_DATA_ROOT))
BASE = DATA_ROOT
ASSEMBLED = BASE / "assembled"


def skip_if_missing_data(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not DATA_ROOT.exists() or len(list(DATA_ROOT.glob("*.tif"))) == 0:
            pytest.skip("Required TIFF files not found.")
        return func(*args, **kwargs)
    return wrapper

@skip_if_missing_data
def test_metadata():
    """Test that metadata can be read from a file."""
    files = mbo.get_files(DATA_ROOT, "tif")
    assert len(files) > 0
    metadata = mbo.get_metadata(files[0])
    assert isinstance(metadata, dict)
    assert "pixel_resolution" in metadata.keys()
    assert "objective_resolution" in metadata.keys()
    assert "dtype" in metadata.keys()
    assert "frame_rate" in metadata.keys()


@skip_if_missing_data
def test_get_files_returns_valid_tiffs():
    files = mbo.get_files(DATA_ROOT, "tif")
    assert isinstance(files, list)
    assert len(files) == 2
    for f in files:
        assert Path(f).suffix in (".tif", ".tiff")
        assert Path(f).exists()


def test_expand_paths(tmp_path):
    """Test expand_paths returns sorted file paths."""
    (tmp_path / "a.txt").write_text("dummy")
    (tmp_path / "b.txt").write_text("dummy")
    (tmp_path / "c.md").write_text("dummy")
    results = mbo.expand_paths(tmp_path)
    names = sorted([Path(p).name for p in results])
    expected = sorted(["a.txt", "b.txt", "c.md"])
    assert names == expected


def test_npy_to_dask(tmp_path):
    """Test npy_to_dask creates a dask array of the expected shape."""
    shape = (10, 20, 30, 40)
    files = []
    for i in range(3):
        arr = np.full(shape, i, dtype=np.float32)
        file_path = tmp_path / f"dummy_{i}.npy"
        np.save(file_path, arr)
        files.append(str(file_path))
    darr = mbo.npy_to_dask(files, name="test", axis=1, astype=np.float32)
    expected_shape = (10, 60, 30, 40)
    assert darr.shape == expected_shape


def test_jupyter_check():
    assert isinstance(mbo.is_running_jupyter(), bool)


def test_imgui_check():
    result = mbo.is_imgui_installed()
    assert isinstance(result, bool)


@skip_if_missing_data
@pytest.mark.parametrize(
    "roi,subdir",
    [
        (0, ""),  # individual ROIs in ASSEMBLED/roi1, roi2…
        (1, ""),  # same, just roi=1
        (None, "full"),  # full‐stack in ASSEMBLED/full
    ],
)
def test_demo_files(roi, subdir):
    ASSEMBLED.mkdir(exist_ok=True)
    files = mbo.get_files(BASE, "tif")
    lazy_array = mbo_imread(files)
    save_dir = ASSEMBLED / subdir if subdir else ASSEMBLED
    lazy_array.roi = roi
    mbo_imwrite(
        lazy_array,
        save_dir,
        ext=".tiff",
        overwrite=True,
        planes=[7, 10],
    )

    out = mbo.get_files(save_dir, "plane", max_depth=2)
    print('------------')
    print(out)


@pytest.fixture
def plane_paths():
    ASSEMBLED.mkdir(exist_ok=True)
    return mbo.get_files(ASSEMBLED, "plane", max_depth=3)


@skip_if_missing_data
def test_full_contains_rois_side_by_side(plane_paths):
    # map parent‐dir → path, e.g. "full", "roi1", "roi2"
    if ASSEMBLED:
        by_dir = {Path(p).parent.name: Path(p) for p in plane_paths}
        full = tfile_imread(by_dir["full"])
        roi1 = tfile_imread(by_dir["roi1"])
        roi2 = tfile_imread(by_dir["roi2"])

        T, H, W = full.shape
        assert roi1.shape == (T, H, W // 2)
        assert roi2.shape == (T, H, W - W // 2)

        left, right = full[:, :, : W // 2], full[:, :, W // 2 :]
        np.testing.assert_array_equal(left, roi1)
        np.testing.assert_array_equal(right, roi2)


@skip_if_missing_data
def test_overwrite_false_skips_existing():
    # First write with overwrite=True
    files = mbo.get_files(BASE, "tif")
    data = mbo_imread(files)
    data.fix_phase = False
    mbo_imwrite(data, ASSEMBLED, ext=".tiff", overwrite=True, planes=[1])
    # Capture output of second call with overwrite=False
    mbo_imwrite(data, ASSEMBLED, ext=".tiff", overwrite=False, planes=[1])
    # Ensure the file exists
    import tifffile
    data = tifffile.imread(ASSEMBLED / "plane1.tif")
    assert data.shape is not None
    print('------------')
    print(data)
    # captured = capsys.readouterr().out
    # assert "All output files exist; skipping save." in captured


@skip_if_missing_data
def test_overwrite_true_rewrites():
    # First write with overwrite=True
    files = mbo.get_files(BASE, "tif")
    data = mbo_imread(files)
    data.fix_phase = False
    mbo_imwrite(data, ASSEMBLED, ext=".tiff", overwrite=True, planes=[1])
    # Capture output of second call with overwrite=True
    mbo_imwrite(data, ASSEMBLED, ext=".tiff", overwrite=True, planes=[1])
    import tifffile
    data = tifffile.imread(ASSEMBLED / "plane1.tif")  # Ensure file exists
    assert data.shape is not None
    print(data)
    # captured = capsys.readouterr().out
    # # And it should print the elapsed‐time message twice (once per call)
    # assert captured.count("Time elapsed:") >= 2

@skip_if_missing_data
def test_benchmark_indexing_test(tmp_path):
    """Benchmark indexing performance for different array types."""
    files = mbo.get_files(BASE, "tif")
    data = mbo_imread(files)

    # Convert to dask and zarr
    dask_array = data.as_dask()
    zarr_array = data.as_zarr()

    arrays = {
        "numpy": data,
        "dask": dask_array,
        "zarr": zarr_array,
    }

    save_path = tmp_path / "benchmark_results.json"
    results = _benchmark_indexing(
        arrays=arrays,
        save_path=save_path,
        num_repeats=3,
        label="Indexing Benchmark",
    )

    assert isinstance(results, dict)
    assert len(results) == 3
    print(results)



import argparse
import concurrent.futures
import sys
import threading
from pathlib import Path

try:
    from fabric import Connection
except ImportError:
    print(
        "Error: Fabric library is not installed. Please install it with 'pip install fabric'.",
        file=sys.stderr,
    )
    sys.exit(1)
from tqdm import tqdm


def parse_args():
    parser = argparse.ArgumentParser(
        description="CLI to sync a local directory to HPC cluster in parallel using Fabric (SFTP over SSH)."
    )
    parser.add_argument(
        "--data",
        type=str,
        required=True,
        help="Local path to the data directory (must be a directory).",
    )
    parser.add_argument(
        "--user", type=str, required=True, help="Username for the remote server."
    )
    parser.add_argument(
        "--remote_host",
        type=str,
        default="dtn02-hpc.rockefeller.edu",
        help="Remote server hostname.",
    )
    parser.add_argument(
        "--remote_path",
        type=str,
        help="Path appended to /lustre/fs4/mbo/scratch/{user}.",
    )
    parser.add_argument(
        "--exclude",
        type=str,
        nargs="*",
        help="List of patterns to exclude (e.g., temp/ or .git/ ).",
    )
    return parser.parse_args()


def validate_paths(data_path, remote_path):
    data_path = Path(data_path).resolve()
    if not data_path.exists() or not data_path.is_dir():
        print(
            f"❌ Error: '{data_path}' must exist and be a directory.", file=sys.stderr
        )
        sys.exit(1)
    if " " in remote_path:
        print(
            f"❌ Error: Remote path '{remote_path}' contains spaces.", file=sys.stderr
        )
        sys.exit(1)
    return data_path, remote_path


def transfer_file(host, user, local_file, remote_file_path, pbar, pbar_lock):
    try:
        file_size = local_file.stat().st_size
        transferred = 0
        # Open a dedicated connection for this file transfer
        with Connection(host=host, user=user) as conn:
            with open(local_file, "rb") as lf, conn.sftp() as sftp:
                remote_dir = "/".join(remote_file_path.split("/")[:-1])
                conn.run(f'mkdir -p "{remote_dir}"')
                with sftp.open(remote_file_path, "wb") as rf:
                    chunk_size = 1024 * 1024
                    while True:
                        chunk = lf.read(chunk_size)
                        if not chunk:
                            break
                        rf.write(chunk)
                        transferred += len(chunk)
                        with pbar_lock:
                            pbar.update(len(chunk))
        if transferred < file_size:
            raise ValueError(
                f"Incomplete transfer for {local_file}: {transferred}/{file_size} bytes."
            )
    except Exception as e:
        print(f"❌ Error transferring {local_file}: {e}", file=sys.stderr)


def transfer_files(host, user, local_path, remote_path, exclude_patterns=None):
    print(f"📂 Starting file transfer from {local_path} to {remote_path}.")
    all_files = [f for f in local_path.rglob("*") if f.is_file()]

    # Filter out excluded files
    if exclude_patterns:

        def should_exclude(f):
            rel_path = str(f.relative_to(local_path)).replace("\\", "/")
            return any(rel_path.startswith(p.rstrip("/")) for p in exclude_patterns)

        all_files = [f for f in all_files if not should_exclude(f)]

    total_files = len(all_files)
    total_size = sum(f.stat().st_size for f in all_files)

    if total_files == 0:
        print("⚠️ No files to transfer. Check exclusions or directory.")
        return

    print(f"📦 Total files: {total_files} | Total size: {total_size / 1e9:.2f} GB")
    pbar_lock = threading.Lock()
    with tqdm(
        total=total_size,
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        desc="Transferring",
    ) as pbar:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for f in all_files:
                relative_path = f.relative_to(local_path)
                remote_file_path = f"{remote_path}/{relative_path}"
                print(f"🚀 Queueing {f} -> {remote_file_path}")
                futures.append(
                    executor.submit(
                        transfer_file, host, user, f, remote_file_path, pbar, pbar_lock
                    )
                )
            concurrent.futures.wait(futures)


def main():
    args = parse_args()

    if not args.remote_path:
        remote_path = f"/lustre/fs4/mbo/scratch/{args.user}"
    else:
        remote_path = (
            f"/lustre/fs4/mbo/scratch/{args.user}/{args.remote_path.strip('/')}"
        )

    local_dir, remote_path = validate_paths(args.data, remote_path)
    transfer_files(
        args.remote_host,
        args.user,
        local_dir,
        remote_path,
        exclude_patterns=args.exclude,
    )
    print("✅ Transfer complete!")


if __name__ == "__main__":
    main()



from __future__ import annotations

import copy
import json
from dataclasses import field
from typing import Any, Tuple, List, Sequence
from dataclasses import dataclass
from pathlib import Path
import numpy as np

import h5py
import tifffile
from dask import array as da
from fsspec.implementations.reference import ReferenceFileSystem
from numpy import ndarray
from zarr import open as zarr_open
from zarr.storage import FsspecStore
import fastplotlib as fpl

from mbo_utilities.metadata import get_metadata
from mbo_utilities._parsing import _make_json_serializable
from mbo_utilities._writers import _write_plane
from mbo_utilities.file_io import (
    _multi_tiff_to_fsspec,
    HAS_ZARR,
    CHUNKS,
    _convert_range_to_slice,
    expand_paths,
)
from mbo_utilities.util import subsample_array
from mbo_utilities.pipelines import HAS_SUITE2P, HAS_MASKNMF, load_from_dir

from mbo_utilities import log
from mbo_utilities.roi import iter_rois
from mbo_utilities.phasecorr import ALL_PHASECORR_METHODS, nd_windowed
from mbo_utilities.scanreader import scans, utils
from mbo_utilities.scanreader.multiroi import ROI

CHUNKS_4D = {0: 1, 1: "auto", 2: -1, 3: -1}
CHUNKS_3D = {0: 1, 1: -1, 2: -1}

logger = log.get("array_types")

@dataclass
class DemixingResultsArray:
    plane_dir: Path

    def load(self):
        data = load_from_dir(self.plane_dir)
        return data["pmd_demixer"].results

    def imshow(self, **kwargs):
        """
        Display the demixing results as an image widget.
        """
        if not HAS_MASKNMF:
            raise ImportError(
                "MaskNMF is not installed. Cannot display demixing results."
            )
        import fastplotlib as fpl

        histogram_widget = kwargs.get("histogram_widget", True)
        figure_kwargs = kwargs.get(
            "figure_kwargs",
            {
                "size": (800, 1000),
            },
        )
        window_funcs = kwargs.get("window_funcs", None)
        return fpl.ImageWidget(
            data=self.load(),
            histogram_widget=histogram_widget,
            figure_kwargs=figure_kwargs,  # "canvas": canvas},
            graphic_kwargs={"vmin": -300, "vmax": 4000},
            window_funcs=window_funcs,
        )


@dataclass
class Suite2pArray:
    metadata: dict

    def __post_init__(self):
        self.Ly = self.metadata["Ly"]
        self.Lx = self.metadata["Lx"]
        self.nframes = self.metadata.get("nframes", self.metadata.get("n_frames"))
        if self.nframes is None:
            raise ValueError("Missing 'nframes' or 'n_frames' in metadata")

        self.filename = self.metadata.get("reg_file", self.metadata.get("raw_file"))
        if self.filename is None:
            raise ValueError("Missing 'reg_file' or 'raw_file' in metadata")

        self.filename = str(self.filename)
        self.shape = (self.nframes, self.Ly, self.Lx)
        self.dtype = np.int16
        self._file = np.memmap(
            self.filename, mode="r", dtype=self.dtype, shape=self.shape
        )

    def __getitem__(self, key):
        return self._file[key]

    def __len__(self):
        return self.shape[0]

    def __array__(self):
        n = min(10, self.nframes)
        return np.stack([self._file[i] for i in range(n)], axis=0)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def min(self):
        return float(self._file[0].min())

    @property
    def max(self):
        return float(self._file[0].max())

    def close(self):
        self._file._mmap.close()


class H5Array:
    def __init__(self, filenames: Path | str, dataset: str = "mov"):
        self.filenames = Path(filenames)
        self._f = h5py.File(self.filenames, "r")
        self._d = self._f[dataset]
        self.shape = self._d.shape
        self.dtype = self._d.dtype
        self.ndim = self._d.ndim

    @property
    def num_planes(self) -> int:
        # TODO: not sure what to do here
        return 14

    def __len__(self) -> int:
        return self.shape[0]

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            key = (key,)

        # Expand ellipsis to match ndim
        if Ellipsis in key:
            idx = key.index(Ellipsis)
            n_missing = self.ndim - (len(key) - 1)
            key = key[:idx] + (slice(None),) * n_missing + key[idx + 1 :]

        # Remove None axes (np.newaxis) and track their positions
        slices = []
        result_shape = []
        dim = 0
        for k in key:
            if k is None:
                result_shape.append(1)
            else:
                slices.append(k)
                dim += 1

        data = self._d[tuple(slices)]

        for i, k in enumerate(key):
            if k is None:
                data = np.expand_dims(data, axis=i)

        return data

    def min(self) -> float:
        return float(self._d[0].min())

    def max(self) -> float:
        return float(self._d[0].max())

    def __array__(self):
        n = min(10, self.shape[0])
        return self._d[:n]

    def close(self):
        self._f.close()

    @property
    def metadata(self) -> dict:
        return dict(self._f.attrs)

    def _imwrite(self, outpath, **kwargs):
        _write_plane(
            self._d, Path(outpath),
            ext=kwargs.get("ext", ".tiff"),
            overwrite=kwargs.get("overwrite", False),
            metadata=self.metadata,
            target_chunk_mb=kwargs.get("target_chunk_mb", 20),
            progress_callback=kwargs.get("progress_callback", None),
            debug=kwargs.get("debug", False),
        )

    def _imwrite2(
        self,
        outpath: Path | str,
        overwrite=False,
        target_chunk_mb=50,
        ext=".tiff",
        progress_callback=None,
        debug=None,
        planes=None,
    ):
        target = outpath
        target.mkdir(exist_ok=True)
        md = self.metadata.copy()
        _write_plane(
            self,
            target,
            overwrite=overwrite,
            ext=ext,
            target_chunk_mb=target_chunk_mb,
            metadata=md,
            progress_callback=progress_callback,
            debug=debug,
        )


@dataclass
class MBOTiffArray:
    filenames: list[Path]
    _chunks: tuple[int, ...] | dict | None = None
    roi: Any = None
    _dask_array: da.Array = field(default=None, init=False, repr=False)

    @property
    def chunks(self) -> tuple[int, ...] | dict:
        return self._chunks or CHUNKS_4D

    @chunks.setter
    def chunks(self, value):
        self._chunks = value

    def _build_dask_array(self) -> da.Array:
        if len(self.filenames) == 1:
            arr = tifffile.memmap(self.filenames[0], mode="r")
            return da.from_array(arr, chunks=self.chunks)

        planes = []
        for p in self.filenames:
            mm = tifffile.memmap(p, mode="r")
            if mm.ndim == 3:
                mm = mm[None, ...]
            planes.append(da.from_array(mm, chunks=self.chunks))

        dstack = da.concatenate(planes, axis=0)  # (Z, T, Y, X)
        return dstack.transpose(1, 0, 2, 3)  # (T, Z, Y, X)

    @property
    def dask(self) -> da.Array:
        if self._dask_array is None:
            self._dask_array = self._build_dask_array()
        return self._dask_array

    def __getitem__(self, key: int | slice | tuple[int, ...]) -> np.ndarray:
        return self.dask[key]

    def __getattr__(self, attr):
        return getattr(self.dask, attr)

    def min(self) -> float:
        return float(self.dask[0].min().compute())

    def max(self) -> float:
        return float(self.dask[0].max().compute())

    @property
    def ndim(self) -> int:
        return self.dask.ndim

    @property
    def shape(self) -> tuple[int, ...]:
        return tuple(self.dask.shape)

    @property
    def metadata(self) -> dict:
        """
        Return metadata from the first TIFF file.
        Assumes all files have the same metadata structure.
        """
        if not self.filenames:
            return {}
        return get_metadata(self.filenames[0])

    def imshow(self, **kwargs) -> fpl.ImageWidget:
        if len(self.filenames) == 1:
            data = tifffile.memmap(self.filenames[0], mode="r")
        else:
            data = self.dask
        histogram_widget = kwargs.get("histogram_widget", True)
        figure_kwargs = kwargs.get(
            "figure_kwargs",
            {
                "size": (800, 1000),
            },
        )
        window_funcs = kwargs.get("window_funcs", None)
        return fpl.ImageWidget(
            data=data,
            histogram_widget=histogram_widget,
            figure_kwargs=figure_kwargs,  # "canvas": canvas},
            graphic_kwargs={"vmin": -300, "vmax": 4000},
            window_funcs=window_funcs,
        )

    def _imwrite(
            self,
            outpath: Path | str,
            overwrite=False,
            target_chunk_mb=50,
            ext=".tiff",
            progress_callback=None,
            debug=None,
            planes=None,
    ):
        if "plane" in self.metadata.keys():
            plane = self.metadata["plane"]
        else:
            raise ValueError("Cannot determine plane from metadata.")

        outpath = Path(outpath)
        ext = ext.lower().lstrip(".")

        if ext in {"bin"}:
            fname = "data_raw.bin"
        else:
            fname = f"plane{plane:03d}.{ext}"

        if outpath.is_dir():
            target = outpath.joinpath(fname)
        else:
            target = outpath.parent.joinpath(fname)

        _write_plane(
            self,
            target,
            overwrite=overwrite,
            target_chunk_mb=target_chunk_mb,
            metadata=self.metadata,
            progress_callback=progress_callback,
            debug=debug,
            dshape=(self.shape[0], self.shape[-1], self.shape[-2]),
            plane_index=None,  # convert to 0-based index
        )


# NOT YET IMPLEMENTED FULLY
@dataclass
class NpyArray:
    filenames: list[Path]

    def load(self) -> Tuple[np.ndarray, List[str]]:
        arr = np.load(str(self.filenames), mmap_mode="r")
        return arr


# NOT YET IMPLEMENTED FULLY
@dataclass
class TiffArray:
    filenames: list[Path]

    def load(self) -> np.memmap | ndarray:
        try:
            return tifffile.memmap(str(self.filenames))
        except (ValueError, MemoryError) as e:
            print(
                f"cannot memmap TIFF file {self.filenames}: {e}\n"
                f" falling back to imread"
            )
            return tifffile.imread(str(self.filenames), mode="r")


class MboRawArray(scans.ScanMultiROI):
    """
    A subclass of ScanMultiROI that ignores the num_fields dimension
    and reorders the output to [time, z, x, y].
    """

    def __init__(
        self,
        files: str | Path | list = None,
        roi: int | Sequence[int] | None = None,
        fix_phase: bool = True,
        phasecorr_method: str = "frame",
        border: int | tuple[int, int, int, int] = 3,
        upsample: int = 5,
        max_offset: int = 4,
    ):
        super().__init__(join_contiguous=True)
        self._metadata = {}  # set when pages are read
        self._fix_phase = fix_phase
        self._phasecorr_method = phasecorr_method
        self.border: int | tuple[int, int, int, int] = border
        self.max_offset: int = max_offset
        self.upsample: int = upsample
        self.use_zarr = False
        self.reference = ""
        self.roi = roi  # alias
        self._roi = roi
        self.pbar = None
        self.show_pbar = False
        self._offset = 0.0

        # Debugging toggles
        self.debug_flags = {
            "frame_idx": True,
            "roi_array_shape": False,
            "phase_offset": False,
        }
        self.logger = logger
        self.logger.info(
            f"Initializing MBO Scan with parameters:\n"
            f"roi: {roi}, "
            f"fix_phase: {fix_phase}, "
            f"phasecorr_method: {phasecorr_method}, "
            f"border: {border}, "
            f"upsample: {upsample}, "
            f"max_offset: {max_offset}"
        )
        if files:
            self.read_data(files)

    def save_fsspec(self, filenames):
        base_dir = Path(filenames[0]).parent

        combined_json_path = base_dir / "combined_refs.json"

        if combined_json_path.is_file():
            # delete it, its cheap to create
            logger.debug(
                f"Removing existing combined reference file: {combined_json_path}"
            )
            combined_json_path.unlink()

        print(f"Generating combined kerchunk reference for {len(filenames)} files…")
        combined_refs = _multi_tiff_to_fsspec(tif_files=filenames, base_dir=base_dir)

        with open(combined_json_path, "w") as _f:
            json.dump(combined_refs, _f)

        print(f"Combined kerchunk reference written to {combined_json_path}")
        self.reference = combined_json_path
        return combined_json_path

    def as_dask(self):
        """
        Convert the current scan data to a Dask array.
        This will create a Dask array in the same directory as the reference file.
        """
        if not HAS_ZARR:
            raise ImportError(
                "Zarr is not installed. Please install it to use this method."
            )
        if not Path(self.reference).is_file():
            raise FileNotFoundError(
                f"Reference file {self.reference} does not exist. "
                "Please call save_fsspec() first."
            )
        return da.from_zarr(
            FsspecStore(ReferenceFileSystem(str(self.reference))),
            chunks=CHUNKS,
        )

    def as_zarr(self):
        """
        Convert the current scan data to a Zarr array.
        This will create a Zarr store in the same directory as the reference file.
        """
        if not HAS_ZARR:
            raise ImportError(
                "Zarr is not installed. Please install it to use this method."
            )
        if not Path(self.reference).is_file():
            return None
        return zarr_open(
            FsspecStore(ReferenceFileSystem(str(self.reference))),
            mode="r",
        )

    def read_data(self, filenames, dtype=np.int16):
        filenames = expand_paths(filenames)
        # try:
        #     self.save_fsspec(filenames)
        #     self.use_zarr = True
        # except Exception as e:
        #     logger.error(f"Failed to save fsspec: {e}")
        self.use_zarr = False
        self.reference = None
        super().read_data(filenames, dtype)
        self._metadata = get_metadata(
            self.tiff_files[0].filehandle.path
        )  # from the file
        self.metadata = (
            {"si": _make_json_serializable(self.tiff_files[0].scanimage_metadata)}
        )
        self._rois = self._create_rois()
        self.fields = self._create_fields()
        if self.join_contiguous:
            self._join_contiguous_fields()

    @property
    def metadata(self):
        md = self._metadata.copy()
        md.update(
            {
                "fix_phase": self.fix_phase,
                "phasecorr_method": self.phasecorr_method,
                "offset": self.offset,
                "border": self.border,
                "upsample": self.upsample,
                "max_offset": self.max_offset,
                "num_frames": self.num_frames,
            }
        )
        return md

    @metadata.setter
    def metadata(self, value):
        self._metadata.update(value)

    @property
    def rois(self):
        """ROI's hold information about the size, position and shape of the ROIs."""
        return self._rois

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, value: float | np.ndarray):
        """
        Set the phase offset for phase correction.
        If value is a scalar, it applies the same offset to all frames.
        If value is an array, it must match the number of frames.
        """
        if isinstance(value, int):
            self._offset = float(value)
        self._offset = value

    @property
    def phasecorr_method(self):
        """
        Get the current phase correction method.
        Options are 'subpix' or 'mean'.
        """
        return self._phasecorr_method

    @phasecorr_method.setter
    def phasecorr_method(self, value: str):
        """
        Set the phase correction method.
        Options are 'two_step', 'subpix', or 'crosscorr'.
        """
        if value not in ALL_PHASECORR_METHODS:
            raise ValueError(
                f"Unsupported phase correction method: {value}. "
                f"Supported methods are: {ALL_PHASECORR_METHODS}"
            )
        self._phasecorr_method = value

    @property
    def fix_phase(self):
        """
        Get whether phase correction is applied.
        If True, phase correction is applied to the data.
        """
        return self._fix_phase

    @fix_phase.setter
    def fix_phase(self, value: bool):
        """
        Set whether to apply phase correction.
        If True, phase correction is applied to the data.
        """
        if not isinstance(value, bool):
            raise ValueError("do_phasecorr must be a boolean value.")
        self._fix_phase = value

    @property
    def roi(self):
        """
        Get the current ROI index.
        If roi is None, returns -1 to indicate no specific ROI.
        """
        return self._roi

    @roi.setter
    def roi(self, value):
        """
        Set the current ROI index.
        If value is None, sets roi to -1 to indicate no specific ROI.
        """
        self._roi = value

    @property
    def num_rois(self) -> int:
        return len(self.rois)

    @property
    def xslices(self):
        return self.fields[0].xslices

    @property
    def yslices(self):
        return self.fields[0].yslices

    @property
    def output_xslices(self):
        return self.fields[0].output_xslices

    @property
    def output_yslices(self):
        return self.fields[0].output_yslices

    def _read_pages(
        self, frames, chans, yslice=slice(None), xslice=slice(None), **kwargs
    ):
        C = self.num_channels
        pages = [f * C + c for f in frames for c in chans]

        H = len(utils.listify_index(yslice, self._page_height))
        W = len(utils.listify_index(xslice, self._page_width))

        if getattr(self, "use_zarr", False):
            zarray = self.as_zarr()
            if zarray is not None:
                buf = np.empty((len(pages), H, W), dtype=self.dtype)
                for i, page in enumerate(pages):
                    f, c = divmod(page, C)
                    buf[i] = zarray[f, c, yslice, xslice]

                if self.fix_phase:
                    self.logger.debug(
                        f"Applying phase correction with strategy: {self.phasecorr_method}"
                    )
                    buf, self.offset = nd_windowed(
                        buf,
                        method=self.phasecorr_method,
                        upsample=self.upsample,
                        max_offset=self.max_offset,
                        border=self.border,
                    )
                return buf.reshape(len(frames), len(chans), H, W)

        # TIFF path
        buf = np.empty((len(pages), H, W), dtype=self.dtype)
        start = 0
        for tf in self.tiff_files:
            end = start + len(tf.pages)
            idxs = [i for i, p in enumerate(pages) if start <= p < end]
            if not idxs:
                start = end
                continue

            frame_idx = [pages[i] - start for i in idxs]
            chunk = tf.asarray(key=frame_idx)[..., yslice, xslice]

            if self.fix_phase:
                self.logger.debug(
                    f"Applying phase correction with strategy: {self.phasecorr_method}"
                )
                corrected, self.offset = nd_windowed(
                    chunk,
                    method=self.phasecorr_method,
                    upsample=self.upsample,
                    max_offset=self.max_offset,
                    border=self.border,
                )
                buf[idxs] = corrected
            else:
                buf[idxs] = chunk
            start = end

        return buf.reshape(len(frames), len(chans), H, W)

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            key = (key,)
        t_key, z_key, _, _ = tuple(_convert_range_to_slice(k) for k in key) + (
            slice(None),
        ) * (4 - len(key))
        frames = utils.listify_index(t_key, self.num_frames)
        chans = utils.listify_index(z_key, self.num_channels)
        if not frames or not chans:
            return np.empty(0)

        logger.debug(
            f"Phase-corrected: {self.fix_phase}/{self.phasecorr_method},"
            f" channels: {chans},"
            f" roi: {self.roi}",
        )
        out = self.process_rois(frames, chans)

        squeeze = []
        if isinstance(t_key, int):
            squeeze.append(0)
        if isinstance(z_key, int):
            squeeze.append(1)
        if squeeze:
            if isinstance(out, tuple):
                out = tuple(np.squeeze(x, axis=tuple(squeeze)) for x in out)
            else:
                out = np.squeeze(out, axis=tuple(squeeze))
        return out

    def process_rois(self, frames, chans):
        if self.roi is not None:
            if isinstance(self.roi, list):
                return tuple(
                    self.process_single_roi(roi_idx - 1, frames, chans)
                    for roi_idx in self.roi
                )
            elif self.roi == 0:
                return tuple(
                    self.process_single_roi(roi_idx, frames, chans)
                    for roi_idx in range(self.num_rois)
                )
            elif isinstance(self.roi, int):
                return self.process_single_roi(self.roi - 1, frames, chans)
        else:
            H_out, W_out = self.field_heights[0], self.field_widths[0]
            out = np.zeros((len(frames), len(chans), H_out, W_out), dtype=self.dtype)
            for roi_idx in range(self.num_rois):
                roi_data = self.process_single_roi(roi_idx, frames, chans)
                oys, oxs = (
                    self.fields[0].output_yslices[roi_idx],
                    self.fields[0].output_xslices[roi_idx],
                )
                out[:, :, oys, oxs] = roi_data
            return out

    def process_single_roi(self, roi_idx, frames, chans):
        return self._read_pages(
            frames,
            chans,
            yslice=self.fields[0].yslices[roi_idx],
            xslice=self.fields[0].xslices[roi_idx],
        )

    @property
    def total_frames(self):
        return sum(len(tf.pages) // self.num_channels for tf in self.tiff_files)

    @property
    def num_planes(self):
        """LBM alias for num_channels."""
        return self.num_channels

    def min(self):
        """
        Returns the minimum value of the first tiff page.
        """
        page = self.tiff_files[0].pages[0]
        return np.min(page.asarray())

    def max(self):
        """
        Returns the maximum value of the first tiff page.
        """
        page = self.tiff_files[0].pages[0]
        return np.max(page.asarray())

    @property
    def shape(self):
        """Shape is relative to the current ROI."""
        if self.roi is not None:
            if not isinstance(self.roi, (list, tuple)):
                if self.roi > 0:
                    s = self.fields[0].output_xslices[self.roi - 1]
                    width = s.stop - s.start
                    return (
                        self.total_frames,
                        self.num_channels,
                        self.field_heights[0],
                        width,
                    )
        # roi = None, or a list/tuple indicates the shape should be relative to the full FOV
        return (
            self.total_frames,
            self.num_channels,
            self.field_heights[0],
            self.field_widths[0],
        )

    @property
    def shape_full(self):
        return (
            self.total_frames,
            self.num_channels,
            self.field_heights[0],
            self.field_widths[0],
        )

    @property
    def ndim(self):
        return 4

    @property
    def size(self):
        return (
            self.num_frames
            * self.num_channels
            * self.field_heights[0]
            * self.field_widths[0]
        )

    @property
    def scanning_depths(self):
        """
        We override this because LBM should always be at a single scanning depth.
        """
        return [0]

    def _create_rois(self):
        """
        Create scan rois from the configuration file. Override the base method to force
        ROI's that have multiple 'zs' to a single depth.
        """
        try:
            roi_infos = self.tiff_files[0].scanimage_metadata["RoiGroups"][
                "imagingRoiGroup"
            ]["rois"]
        except KeyError:
            raise RuntimeError(
                "This file is not a raw-scanimage tiff or is missing tiff.scanimage_metadata."
            )
        roi_infos = roi_infos if isinstance(roi_infos, list) else [roi_infos]

        # discard empty/malformed ROIs
        roi_infos = list(
            filter(lambda r: isinstance(r["zs"], (int, float, list)), roi_infos)
        )

        # LBM uses a single depth that is not stored in metadata,
        # so force this to be 0.
        for roi_info in roi_infos:
            roi_info["zs"] = [0]

        rois = [ROI(roi_info) for roi_info in roi_infos]
        return rois

    def _create_fields(self):
        """Go over each slice depth and each roi generating the scanned fields."""
        fields = []
        previous_lines = 0
        for slice_id, scanning_depth in enumerate(self.scanning_depths):
            next_line_in_page = 0  # each slice is one tiff page
            for roi_id, roi in enumerate(self.rois):
                new_field = roi.get_field_at(scanning_depth)
                if new_field is not None:
                    print(f"Scanning depth {scanning_depth}, ROI {roi_id} ")
                    # if next_line_in_page + new_field.height > self._page_height:
                    #     error_msg = (
                    #         "Overestimated number of fly to lines ({}) at "
                    #         "scanning depth {}".format(
                    #             self._num_fly_to_lines, scanning_depth
                    #         )
                    #     )
                    #     raise RuntimeError(error_msg)

                    # Set xslice and yslice (from where in the page to cut it)
                    new_field.yslices = [
                        slice(next_line_in_page, next_line_in_page + new_field.height)
                    ]
                    new_field.xslices = [slice(0, new_field.width)]

                    # Set output xslice and yslice (where to paste it in output)
                    new_field.output_yslices = [slice(0, new_field.height)]
                    new_field.output_xslices = [slice(0, new_field.width)]

                    # Set slice and roi id
                    new_field.slice_id = slice_id
                    new_field.roi_ids = [roi_id]

                    offsets = self._compute_offsets(
                        new_field.height, previous_lines + next_line_in_page
                    )
                    new_field.offsets = [offsets]
                    next_line_in_page += new_field.height + self._num_fly_to_lines
                    fields.append(new_field)
            previous_lines += self._num_lines_between_fields
        return fields


    def __array__(self):
        """
        Convert the scan data to a NumPy array.
        Calculate the size of the scan and subsample to keep under memory limits.
        """
        return subsample_array(self, ignore_dims=[-1, -2, -3])

    def _imwrite(
        self,
        outpath: Path | str,
        overwrite=False,
        target_chunk_mb=50,
        ext=".tiff",
        progress_callback=None,
        debug=None,
        planes=None,
    ):
        # convert to 0 based indexing
        if isinstance(planes, int):
            planes = [planes - 1]
        elif planes is None:
            planes = list(range(self.num_planes))
        else:
            planes = [p - 1 for p in planes]
        for roi in iter_rois(self):
            for plane in planes:
                self.roi = roi
                if roi is None:
                    fname = f"plane{plane+1:02d}_stitched{ext}"
                else:
                    fname = f"plane{plane+1:02d}_roi{roi}{ext}"
                target = outpath.joinpath(fname)
                target.parent.mkdir(exist_ok=True)

                md = self.metadata.copy()
                md["plane"] = plane + 1  # back to 1-based indexing
                md["roi"] = roi
                _write_plane(
                    self,
                    target,
                    overwrite=overwrite,
                    target_chunk_mb=target_chunk_mb,
                    metadata=md,
                    progress_callback=progress_callback,
                    debug=debug,
                    dshape=(self.shape[0], self.shape[-1], self.shape[-2]),
                    plane_index=plane,
                )

    def imshow(self, **kwargs):
        arrays = []
        names = []
        # if roi is None, use a single array.roi = None
        # if roi is 0, get a list of all ROIs by deeepcopying the array and setting each roi
        for roi in iter_rois(self):
            arr = copy.copy(self)
            arr.roi = roi
            arrays.append(arr)
            names.append(f"ROI {roi}" if roi else "Full Image")

        figure_shape = (1, len(arrays))

        histogram_widget = kwargs.get("histogram_widget", True)
        figure_kwargs = kwargs.get(
            "figure_kwargs",
            {
                "size": (800, 1000),
            },
        )
        window_funcs = kwargs.get("window_funcs", None)
        return fpl.ImageWidget(
            data=arrays,
            names=names,
            histogram_widget=histogram_widget,
            figure_kwargs=figure_kwargs,  # "canvas": canvas},
            figure_shape=figure_shape,
            graphic_kwargs={"vmin": -300, "vmax": 4000},
            window_funcs=window_funcs,
        )



import logging
import webbrowser
from pathlib import Path
from typing import Literal
import threading
from functools import partial
from dataclasses import dataclass, field

import numpy as np
from numpy import ndarray
from scipy.ndimage import gaussian_filter
from skimage.registration import phase_cross_correlation

from imgui_bundle import (
    imgui,
    hello_imgui,
    imgui_ctx,
    implot,
    portable_file_dialogs as pfd,
    ImVec2,
)

from mbo_utilities.file_io import (
    MBO_SUPPORTED_FTYPES,
    get_mbo_dirs,
)
from mbo_utilities.array_types import MboRawArray
from mbo_utilities.graphics._imgui import (
    begin_popup_size,
    ndim_to_frame,
    style_seaborn_dark,
)
from mbo_utilities.graphics._widgets import (
    set_tooltip,
    checkbox_with_tooltip,
    draw_scope,
)
from mbo_utilities.graphics.progress_bar import (
    draw_zstats_progress,
    draw_saveas_progress,
)
from mbo_utilities.graphics.pipeline_widgets import Suite2pSettings, draw_tab_process
from mbo_utilities.lazy_array import imread, imwrite
from mbo_utilities.phasecorr import apply_scan_phase_offsets
from mbo_utilities.graphics.gui_logger import GuiLogger, GuiLogHandler, GUI_LOGGERS
from mbo_utilities import log

try:
    import cupy as cp  # noqa
    from cusignal import (
        register_translation,
    )  # GPU version of phase_cross_correlation # noqa

    HAS_CUPY = True
except ImportError:
    HAS_CUPY = False
    register_translation = phase_cross_correlation  # noqa

try:
    import torch

    HAS_TORCH = True
except ImportError:
    HAS_TORCH = False
    torch = None

import fastplotlib as fpl
from fastplotlib.ui import EdgeWindow

REGION_TYPES = ["Full FOV", "Sub-FOV"]
USER_PIPELINES = ["suite2p", "masknmf"]


def _save_as_worker(path, **imwrite_kwargs):
    data = imread(path, roi=imwrite_kwargs.pop("roi", None))
    imwrite(data, **imwrite_kwargs)


@dataclass
class SaveStatus:
    # progressbar
    progress: float = 0.0
    plane: int | None = None
    message: str = ""
    logs: dict = field(default_factory=dict)


def draw_menu(parent):
    # (accessible from the "Tools" menu)
    if parent.show_style_window:
        _, parent.show_style_window = imgui.begin(
            "Style Editor", parent.show_style_window
        )
        imgui.show_style_editor()
        imgui.end()
    if parent.show_scope_window:
        size = begin_popup_size()
        imgui.set_next_window_size(size, imgui.Cond_.first_use_ever)  # type: ignore # noqa
        _, parent.show_scope_window = imgui.begin(
            "Scope Inspector",
            parent.show_scope_window,
        )
        draw_scope()
        imgui.end()
    if parent.show_debug_panel:
        size = begin_popup_size()
        imgui.set_next_window_size(size, imgui.Cond_.first_use_ever)  # type: ignore # noqa
        _, parent.show_debug_panel = imgui.begin(
            "MBO Debug Panel",
            parent.show_debug_panel,
        )
        parent.debug_panel.draw()
        imgui.end()
    with imgui_ctx.begin_child(
        "menu",
        window_flags=imgui.WindowFlags_.menu_bar,  # noqa,
        child_flags=imgui.ChildFlags_.auto_resize_y
        | imgui.ChildFlags_.always_auto_resize,
    ):
        if imgui.begin_menu_bar():
            if imgui.begin_menu("File", True):
                if imgui.menu_item(
                    "Save as", "Ctrl+S", p_selected=False, enabled=parent.is_mbo_scan
                )[0]:
                    parent._saveas_popup_open = True
                imgui.end_menu()
            if imgui.begin_menu("Docs", True):
                if imgui.menu_item(
                    "Open Docs", "Ctrl+I", p_selected=False, enabled=True
                )[0]:
                    webbrowser.open(
                        "https://millerbrainobservatory.github.io/mbo_utilities/"
                    )
                imgui.end_menu()
            if imgui.begin_menu("Settings", True):
                imgui.text_colored(imgui.ImVec4(0.8, 1.0, 0.2, 1.0), "Tools")
                imgui.separator()
                imgui.spacing()
                _, parent.show_style_window = imgui.menu_item(
                    "Style Editor", "", parent.show_style_window, True
                )
                _, parent.show_debug_panel = imgui.menu_item(
                    "Debug Panel",
                    "",
                    p_selected=parent.show_debug_panel,
                    enabled=True,
                )
                _, parent.show_scope_window = imgui.menu_item(
                    "Scope Inspector", "", parent.show_scope_window, True
                )
                imgui.end_menu()
        imgui.end_menu_bar()
    pass


def draw_tabs(parent):
    with imgui_ctx.begin_child(
        "tabs",
    ):
        if imgui.begin_tab_bar("MainPreviewTabs"):
            if imgui.begin_tab_item("Preview")[0]:
                imgui.push_style_var(imgui.StyleVar_.window_padding, imgui.ImVec2(0, 0))  # noqa
                imgui.push_style_var(imgui.StyleVar_.frame_padding, imgui.ImVec2(0, 0))  # noqa
                parent.draw_preview_section()
                imgui.pop_style_var()
                imgui.pop_style_var()
                imgui.end_tab_item()
            if imgui.begin_tab_item("Summary Stats")[0]:
                imgui.push_style_var(imgui.StyleVar_.window_padding, imgui.ImVec2(0, 0))  # noqa
                imgui.push_style_var(imgui.StyleVar_.frame_padding, imgui.ImVec2(0, 0))  # noqa
                parent.draw_stats_section()
                imgui.pop_style_var()
                imgui.pop_style_var()
                imgui.end_tab_item()
            if imgui.begin_tab_item("Process")[0]:
                draw_tab_process(parent)
                imgui.end_tab_item()
            imgui.end_tab_bar()


def draw_saveas_popup(parent):
    if getattr(parent, "_saveas_popup_open"):
        imgui.open_popup("Save As")
        parent._saveas_popup_open = False

    if imgui.begin_popup_modal("Save As")[0]:
        imgui.dummy(ImVec2(0, 5))

        imgui.set_next_item_width(hello_imgui.em_size(25))

        # Directory + Ext
        _, parent._saveas_outdir = imgui.input_text(
            "Save Dir",
            str(Path(parent._saveas_outdir).expanduser().resolve()),
            256,
        )
        imgui.same_line()
        if imgui.button("Browse"):
            home = Path().home()
            res = pfd.select_folder(str(home))
            if res:
                parent._saveas_outdir = res.result()

        imgui.set_next_item_width(hello_imgui.em_size(25))
        _, parent._ext_idx = imgui.combo("Ext", parent._ext_idx, MBO_SUPPORTED_FTYPES)
        parent._ext = MBO_SUPPORTED_FTYPES[parent._ext_idx]

        imgui.spacing()
        imgui.separator()
        imgui.spacing()

        # Options Section
        parent._saveas_rois = checkbox_with_tooltip(
            "Save ROI's",
            parent._saveas_rois,
            "Enable to save each ROI individually. Saved to subfolders like roi1/, roi2/, etc.",
        )
        if parent._saveas_rois:
            try:
                num_rois = parent.image_widget.data[0].num_rois
            except Exception as e:
                num_rois = 1

            imgui.spacing()
            imgui.separator()
            imgui.text_colored(imgui.ImVec4(0.8, 0.8, 0.2, 1.0), "Choose ROI(s):")
            imgui.dummy(ImVec2(0, 5))

            if imgui.button("All##roi"):
                parent._saveas_selected_roi = set(range(num_rois))
            imgui.same_line()
            if imgui.button("None##roi"):
                parent._saveas_selected_roi = set()

            imgui.columns(2, borders=False)
            for i in range(num_rois):
                imgui.push_id(f"roi_{i}")
                selected = i in parent._saveas_selected_roi
                _, selected = imgui.checkbox(f"ROI {i + 1}", selected)
                if selected:
                    parent._saveas_selected_roi.add(i)
                else:
                    parent._saveas_selected_roi.discard(i)
                imgui.pop_id()
                imgui.next_column()
            imgui.columns(1)

        imgui.spacing()
        imgui.separator()

        imgui.text_colored(imgui.ImVec4(0.8, 0.8, 0.2, 1.0), "Options")
        set_tooltip(
            "Note: Current values for upsample and max-offset are applied during scan-phase correction.",
            True,
        )

        imgui.dummy(ImVec2(0, 5))

        parent._overwrite = checkbox_with_tooltip(
            "Overwrite", parent._overwrite, "Replace any existing output files."
        )

        fix_phase_changed, fix_phase_value = imgui.checkbox(
            "Fix Scan Phase", parent._fix_phase
        )
        imgui.same_line()
        imgui.text_disabled("(?)")
        if imgui.is_item_hovered():
            imgui.begin_tooltip()
            imgui.push_text_wrap_pos(imgui.get_font_size() * 35.0)
            imgui.text_unformatted("Apply scan-phase correction to interleaved lines.")
            imgui.pop_text_wrap_pos()
            imgui.end_tooltip()
        if fix_phase_changed:
            parent.fix_phase = fix_phase_value

        parent._debug = checkbox_with_tooltip(
            "Debug",
            parent._debug,
            "Run with debugging, settings -> debug to view the outputs.",
        )
        parent._saveas_save_phase_png = checkbox_with_tooltip(
            "Save Phase Images",
            parent._saveas_save_phase_png,
            "Saves pre-post scan-phase images as PNGs to the save-directory.",
        )

        imgui.spacing()
        imgui.text("Chunk Size (MB)")
        set_tooltip(
            "Target chunk size when saving TIFF or binary. Affects I/O and memory usage."
        )

        imgui.set_next_item_width(hello_imgui.em_size(20))
        _, parent._saveas_chunk_mb = imgui.drag_int(
            "##target_chunk_mb",
            parent._saveas_chunk_mb,
            v_speed=1,
            v_min=1,
            v_max=1024,
        )

        imgui.spacing()
        imgui.separator()

        # Z-plane selection
        imgui.text_colored(imgui.ImVec4(0.8, 0.8, 0.2, 1.0), "Choose z-planes:")
        imgui.dummy(ImVec2(0, 5))

        try:
            num_planes = parent.image_widget.data[0].num_channels  # noqa
        except Exception as e:
            num_planes = 1
            hello_imgui.log(
                hello_imgui.LogLevel.error,
                f"Could not read number of planes: {e}",
            )

        if imgui.button("All"):
            parent._selected_planes = set(range(num_planes))
        imgui.same_line()
        if imgui.button("None"):
            parent._selected_planes = set()

        imgui.columns(2, borders=False)
        for i in range(num_planes):
            imgui.push_id(i)
            selected = i in parent._selected_planes
            _, selected = imgui.checkbox(f"Plane {i + 1}", selected)
            if selected:
                parent._selected_planes.add(i)
            else:
                parent._selected_planes.discard(i)
            imgui.pop_id()
            imgui.next_column()
        imgui.columns(1)

        imgui.spacing()
        imgui.separator()
        imgui.spacing()

        if imgui.button("Save", imgui.ImVec2(100, 0)):
            if not parent._saveas_outdir:
                parent._saveas_outdir = get_mbo_dirs()["data"].joinpath("data")
            try:
                save_planes = [p + 1 for p in parent._selected_planes]
                parent._saveas_total = len(save_planes)
                if parent._saveas_rois:
                    if (
                        not parent._saveas_selected_roi
                        or len(parent._saveas_selected_roi) == set()
                    ):
                        parent._saveas_selected_roi = set(
                            range(1, parent.num_arrays + 1)
                        )
                    rois = sorted(parent._saveas_selected_roi)
                else:
                    rois = None

                save_kwargs = {
                    "path": parent.fpath,
                    "outpath": parent._saveas_outdir,
                    "planes": save_planes,
                    "roi": rois,
                    "overwrite": parent._overwrite,
                    "debug": parent._debug,
                    "ext": parent._ext,
                    "target_chunk_mb": parent._saveas_chunk_mb,
                    "progress_callback": lambda frac,
                    current_plane: parent.gui_progress_callback(frac, current_plane),
                }
                parent.logger.info(f"Saving planes {save_planes}")
                parent.logger.info(
                    f"Saving to {parent._saveas_outdir} as {parent._ext}"
                )
                threading.Thread(
                    target=_save_as_worker, kwargs=save_kwargs, daemon=True
                ).start()
                imgui.close_current_popup()
            except Exception as e:
                parent.logger.info(f"Error saving data: {e}")
                imgui.close_current_popup()

        imgui.same_line()
        if imgui.button("Cancel"):
            imgui.close_current_popup()

        imgui.end_popup()


class PreviewDataWidget(EdgeWindow):
    def __init__(
        self,
        iw: fpl.ImageWidget,
        fpath: str | None | list = None,
        threading_enabled: bool = True,
        size: int = None,
        location: Literal["bottom", "right"] = "right",
        title: str = "Data Preview",
        show_title: bool = False,
        movable: bool = False,
        resizable: bool = False,
        scrollable: bool = False,
        auto_resize: bool = True,
        window_flags: int | None = None,
        rois: int = 1,
    ):
        """
        Fastplotlib attachment, callable with fastplotlib.ImageWidget.add_gui(PreviewDataWidget)
        """
        self.debug_panel = GuiLogger()
        gui_handler = GuiLogHandler(self.debug_panel)
        for name in GUI_LOGGERS:
            lg = log.get(name)
            lg.addHandler(gui_handler)
            lg.setLevel(logging.WARNING)
            lg.disabled = False
            lg.propagate = True
        self.logger = log.get("gui")
        self.s2p = Suite2pSettings()
        self.logger.info("Logger initialized.")

        flags = (
            (imgui.WindowFlags_.no_title_bar if not show_title else 0)
            | (imgui.WindowFlags_.no_move if not movable else 0)
            | (imgui.WindowFlags_.no_resize if not resizable else 0)
            | (imgui.WindowFlags_.no_scrollbar if not scrollable else 0)
            | (imgui.WindowFlags_.always_auto_resize if auto_resize else 0)
            | (window_flags or 0)
        )
        super().__init__(
            figure=iw.figure,
            size=350 if size is None else size,
            location=location,
            title=title,
            window_flags=flags,
        )
        if implot.get_current_context() is None:
            implot.create_context()

        # backend.create_fonts_texture()
        io = imgui.get_io()
        fd_settings_dir = (
            Path(get_mbo_dirs()["imgui"])
            .joinpath("assets", "app_settings", "preview_settings.ini")
            .expanduser()
            .resolve()
        )
        io.set_ini_filename(str(fd_settings_dir))

        fonts_path = Path(get_mbo_dirs()["assets"]).joinpath("fonts")
        self._fonts = []
        for font_file in fonts_path.rglob("*.ttf"):
            self._fonts.append(io.fonts.add_font_from_file_ttf(str(font_file), 16))
        io.fonts.build()

        self.font_size = 12
        self.fpath = fpath if fpath else getattr(iw, "fpath", None)

        # image widget setup
        self.image_widget = iw
        self.rois = rois
        self.shape = self.image_widget.data[0].shape
        if self.image_widget.window_funcs is None:
            self.image_widget.window_funcs = {"t": (np.mean, 0)}

        if len(self.shape) == 4:
            self.nz = self.shape[1]
        elif len(self.shape) == 3:
            self.nz = 1
        else:
            self.nz = 1

        self.is_mbo_scan = (
            True if isinstance(self.image_widget.data[0], MboRawArray) else False
        )
        if self.is_mbo_scan:
            for arr in self.image_widget.data:
                arr.fix_phase = False

        for subplot in self.image_widget.figure:
            subplot.toolbar = False

        self.image_widget._image_widget_sliders._loop = True  # noqa
        if hasattr(self.image_widget, "rois") or hasattr(
            self.image_widget.data[0], "rois"
        ):
            self._array_type = "roi"
        else:
            self._array_type = "array"

        # boolean flags: imgui.begin() calls that are drawn
        # when these are set to true.
        self.show_style_window = False
        self.show_metrics_window = False
        self.show_debug_panel = False
        self.show_scope_window = False

        # properties: each have a @property getter and setter
        # that controls what to do with the value.
        # different filetypes sometimes require different handling
        # though it may be easier to group these instead into a function or a class.
        self._max_offset = 3
        self._gaussian_sigma = 0
        self._current_offset = [0.0] * self.num_arrays
        self._window_size = 1
        self._phase_upsample = 20
        self._border = 3
        self._auto_update = False
        self._proj = "mean"

        self._selected_pipelines = None
        self._selected_array = 0

        # properties for saving to another filetype
        self._ext = str(getattr(self, "_ext", ".tiff"))
        self._ext_idx = MBO_SUPPORTED_FTYPES.index(".tiff")
        self._selected_planes = set()
        self._planes_str = str(getattr(self, "_planes_str", ""))
        self._overwrite = True
        self._fix_phase = False
        self._debug = False
        self._saveas_save_phase_png = False
        self._saveas_chunk_mb = 20

        # zstats: an entry for each given array
        # these are sent to a thread to compute
        self._zstats = [
            {"mean": [], "std": [], "snr": []} for _ in range(self.num_arrays)
        ]
        self._zstats_means = [0 for _ in range(self.num_arrays)]
        self._zstats_done = [False] * self.num_arrays
        self._zstats_progress = [0.0] * self.num_arrays
        self._zstats_current_z = [0] * self.num_arrays

        self._saveas_popup_open = False
        self._saveas_done = False
        self._saveas_progress = 0.0
        self._saveas_current_index = 0
        self._saveas_outdir = str(getattr(self, "_save_dir", ""))
        self._saveas_total = 0

        self._saveas_selected_roi = set()  # -1 means all ROIs
        self._saveas_rois = False
        self._saveas_selected_roi_mode = "All"
        self.set_context_info()

        if threading_enabled:
            self.logger.info("Starting zstats computation in a separate thread.")
            threading.Thread(target=self.compute_zstats, daemon=True).start()

    def set_context_info(self):
        if isinstance(self.fpath, list):
            title = f"{[Path(f).stem for f in self.fpath]}"
        else:
            title = f"Filepath: {Path(self.fpath).stem}"
        self.image_widget.figure.canvas.set_title(str(title))

    @property
    def current_offset(self) -> list[float]:
        if not self.fix_phase:
            return [0.0 for _ in self.image_widget.data]
        if not self.is_mbo_scan:
            self.logger.critical(
                "Scan-phase correction is only implemented for MBO scans. "
                "Returning zero offsets."
            )
            return [0.0 for _ in self.image_widget.data]
        if all(hasattr(array, "offset") for array in self.image_widget.data):
            self.logger.debug(
                f"All arrays have offset attribute. Setting from array.offset"
            )
            return [array.offset for array in self.image_widget.data]
        else:
            raise NotImplementedError(
                "Scan-phase correction is only implemented for MBO scans."
            )

    @property
    def fix_phase(self):
        return self._fix_phase

    @fix_phase.setter
    def fix_phase(self, value):
        self._fix_phase = value
        if self.is_mbo_scan:
            for arr in self.image_widget.data:
                if isinstance(arr, MboRawArray):
                    arr.fix_phase = value
        else:
            self.update_frame_apply()

        # force update
        self.image_widget.current_index = self.image_widget.current_index

    @property
    def border(self):
        return self._border

    @border.setter
    def border(self, value):
        self._border = value
        for arr in self.image_widget.data:
            if isinstance(arr, MboRawArray):
                arr.border = value
                self.logger.info(f"Border set to {value}.")
            else:
                self.logger.warning(
                    "Max offset is only applicable to MBO Scan objects. "
                    "No action taken."
                )

    @property
    def max_offset(self):
        return self._max_offset

    @max_offset.setter
    def max_offset(self, value):
        self._max_offset = value
        for arr in self.image_widget.data:
            if isinstance(arr, MboRawArray):
                arr.max_offset = value
                self.logger.info(f"Max offset set to {value}.")
            else:
                self.logger.warning(
                    "Max offset is only applicable to MBO Scan objects. "
                    "No action taken."
                )

    @property
    def num_arrays(self):
        return len(self.image_widget.managed_graphics)

    @property
    def selected_array(self):
        return self._selected_array

    @selected_array.setter
    def selected_array(self, value):
        if value < 0 or value >= len(self.image_widget.data):
            raise ValueError(
                f"Invalid array index: {value}. "
                f"Must be between 0 and {len(self.image_widget.managed_graphics) - 1}."
            )
        self._selected_array = value
        self.logger.info(f"Selected array index set to {value}.")
        # self.image_widget.current_index = {"roi": value}
        self.update_frame_apply()

    @property
    def gaussian_sigma(self):
        return self._gaussian_sigma

    @gaussian_sigma.setter
    def gaussian_sigma(self, value):
        if value > 0:
            self._gaussian_sigma = value
            self.logger.info(f"Gaussian sigma set to {value}.")
            self.update_frame_apply()
        else:
            self.logger.warning(f"Invalid gaussian sigma value: {value}. ")

    @property
    def proj(self):
        return self._proj

    @proj.setter
    def proj(self, value):
        if value != self._proj:
            if value == "mean-sub":
                self.logger.info("Setting projection to mean-subtracted.")
                self.update_frame_apply()
            else:
                self.logger.info(f"Setting projection to np.{value}.")
                self.image_widget.window_funcs["t"].func = getattr(np, value)
            self._proj = value

    @property
    def window_size(self):
        return self._window_size

    @window_size.setter
    def window_size(self, value):
        self.image_widget.window_funcs["t"].window_size = value
        self._window_size = value
        self.logger.info(f"Window size set to {value}.")

    @property
    def phase_upsample(self):
        return self._phase_upsample

    @phase_upsample.setter
    def phase_upsample(self, value):
        self._phase_upsample = value
        if self.is_mbo_scan:
            for arr in self.image_widget.data:
                if isinstance(arr, MboRawArray):
                    arr.upsample = value
        else:
            self.logger.warning(
                "Phase upsample is only applicable to MBO Scan objects. "
                "No action taken."
            )

    def update(self):
        draw_saveas_popup(self)
        draw_menu(self)
        draw_tabs(self)

    def draw_stats_section(self):
        if not self._zstats_done:
            return

        stats_list = self._zstats if isinstance(self._zstats, list) else [self._zstats]

        imgui.text_colored(imgui.ImVec4(0.8, 1.0, 0.2, 1.0), "Z-Plane Summary Stats")
        cflags = imgui.ChildFlags_.auto_resize_y | imgui.ChildFlags_.always_auto_resize  # type: ignore # noqa
        imgui.spacing()

        # ROI selector
        array_labels = [
            f"{self._array_type} {i + 1}"
            for i in range(len(stats_list))
            if stats_list[i] and "mean" in stats_list[i]
        ]
        array_labels.append("Combined")
        avail = imgui.get_content_region_avail().x
        xpos = 0

        for i, label in enumerate(array_labels):
            if imgui.radio_button(label, self._selected_array == i):
                self._selected_array = i
            button_width = (
                imgui.calc_text_size(label).x + imgui.get_style().frame_padding.x * 4
            )
            xpos += button_width + imgui.get_style().item_spacing.x

            if xpos >= avail:
                xpos = button_width
                imgui.new_line()
            else:
                imgui.same_line()

        imgui.separator()

        if self._selected_array == len(array_labels) - 1:  # Combined
            imgui.text(f"Stats for Combined {self._array_type}s")
            mean_vals = np.mean(
                [np.array(s["mean"]) for s in stats_list if s and "mean" in s], axis=0
            )

            if len(mean_vals) == 0:
                return

            std_vals = np.mean(
                [np.array(s["std"]) for s in stats_list if s and "std" in s], axis=0
            )
            snr_vals = np.mean(
                [np.array(s["snr"]) for s in stats_list if s and "snr" in s], axis=0
            )

            z_vals = np.arange(1, len(mean_vals) + 1, dtype=np.float32)

            # Table
            with imgui_ctx.begin_child(
                "##SummaryCombined", size=imgui.ImVec2(0, 0), child_flags=cflags
            ):
                if imgui.begin_table(
                    f"Stats, averaged over {self._array_type}s",
                    4,
                    imgui.TableFlags_.borders | imgui.TableFlags_.row_bg,  # type: ignore # noqa
                ):  # type: ignore # noqa
                    for col in ["Z", "Mean", "Std", "SNR"]:
                        imgui.table_setup_column(
                            col, imgui.TableColumnFlags_.width_stretch
                        )  # type: ignore # noqa
                    imgui.table_headers_row()
                    for i in range(len(z_vals)):
                        imgui.table_next_row()
                        for val in (z_vals[i], mean_vals[i], std_vals[i], snr_vals[i]):
                            imgui.table_next_column()
                            imgui.text(f"{val:.2f}")
                    imgui.end_table()

            with imgui_ctx.begin_child(
                "##PlotsCombined", size=imgui.ImVec2(0, 0), child_flags=cflags
            ):
                imgui.text("Z-plane Signal: Combined")
                if implot.begin_plot("Z-Plane Plot", ImVec2(-1, 300)):
                    implot.setup_axes(
                        "Z-Plane",
                        "Mean Fluorescence",
                        implot.AxisFlags_.none.value,
                        implot.AxisFlags_.auto_fit.value,
                    )
                    implot.setup_axis_limits(implot.ImAxis_.x1.value, 1, self.nz)
                    implot.setup_axis_format(implot.ImAxis_.x1.value, "%g")
                    style_seaborn_dark()
                    if len(z_vals) == 1:
                        implot.plot_bars("Mean", z_vals, mean_vals, 0.5)
                        implot.plot_error_bars("Std", z_vals, mean_vals, std_vals)
                    else:
                        implot.plot_error_bars(
                            "Mean ± Std", z_vals, mean_vals, std_vals
                        )
                        implot.plot_line("Mean", z_vals, mean_vals)
                    implot.end_plot()

        else:
            array_idx = self._selected_array
            stats = stats_list[array_idx]

            if not stats or "mean" not in stats:
                return

            imgui.text(f"Stats for {self._array_type} {array_idx + 1}")
            mean_vals = np.array(stats["mean"])
            std_vals = np.array(stats["std"])
            snr_vals = np.array(stats["snr"])
            z_vals = np.arange(1, len(mean_vals) + 1, dtype=np.float32)

            with imgui_ctx.begin_child(
                f"##Summary{array_idx}", size=imgui.ImVec2(0, 0), child_flags=cflags
            ):
                if imgui.begin_table(
                    f"zstats{array_idx}",
                    4,
                    imgui.TableFlags_.borders | imgui.TableFlags_.row_bg,
                ):  # type: ignore # noqa
                    for col in ["Z", "Mean", "Std", "SNR"]:
                        imgui.table_setup_column(
                            col, imgui.TableColumnFlags_.width_stretch
                        )  # type: ignore # noqa
                    imgui.table_headers_row()
                    for i in range(len(z_vals)):
                        imgui.table_next_row()
                        for val in (z_vals[i], mean_vals[i], std_vals[i], snr_vals[i]):
                            imgui.table_next_column()
                            imgui.text(f"{val:.2f}")
                    imgui.end_table()

            with imgui_ctx.begin_child(
                f"##Plots", size=imgui.ImVec2(0, 0), child_flags=cflags
            ):
                imgui.text("Z-plane Signal: Mean ± Std")
                if implot.begin_plot(
                    f"Z-Plane Signal {array_idx}", size=imgui.ImVec2(-1, 300)
                ):
                    z_vals = np.arange(1, len(mean_vals) + 1, dtype=np.float32)
                    implot.setup_axes(
                        "Z-Plane",
                        "Mean Fluorescence",
                        implot.AxisFlags_.none.value,
                        implot.AxisFlags_.auto_fit.value,
                    )
                    implot.setup_axis_limits(implot.ImAxis_.x1.value, 1, self.nz)
                    implot.setup_axis_format(implot.ImAxis_.x1.value, "%g")
                    style_seaborn_dark()
                    if len(z_vals) == 1:
                        implot.plot_bars("Mean", z_vals, mean_vals, 0.5)
                        implot.plot_error_bars("Std", z_vals, mean_vals, std_vals)
                    else:
                        implot.plot_error_bars(
                            "Mean ± Std", z_vals, mean_vals, std_vals
                        )
                        implot.plot_line("Mean", z_vals, mean_vals)
                    implot.end_plot()

    def draw_preview_section(self):
        imgui.dummy(ImVec2(0, 5))
        cflags = imgui.ChildFlags_.auto_resize_y | imgui.ChildFlags_.always_auto_resize  # noqa
        with imgui_ctx.begin_child(
            "##PreviewChild",
            imgui.ImVec2(0, 0),
            cflags,
        ):
            # Section: Window Functions
            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            imgui.text_colored(imgui.ImVec4(0.8, 0.8, 0.2, 1.0), "Window Functions")
            imgui.spacing()

            imgui.push_style_var(imgui.StyleVar_.frame_padding, imgui.ImVec2(2, 2))  # noqa

            imgui.begin_group()
            options = ["mean", "max", "std"]
            disabled_label = (
                "mean-sub (pending)" if not self._zstats_done else "mean-sub"
            )
            options.append(disabled_label)

            current_display_idx = options.index(
                self.proj if self._proj != "mean-sub" else disabled_label
            )

            imgui.set_next_item_width(hello_imgui.em_size(15))
            proj_changed, selected_display_idx = imgui.combo(
                "Projection", current_display_idx, options
            )
            set_tooltip(
                "Choose projection method over the sliding window: “mean” (average), “max” (peak), “std” (variance), or “mean-sub” (background-subtracted mean, recommended for motion preview)."
            )

            if proj_changed:
                selected_label = options[selected_display_idx]
                if selected_label == "mean-sub (pending)":
                    pass  # ignore user click while disabled
                else:
                    self.proj = selected_label
                    if self.proj == "mean-sub":
                        self.update_frame_apply()
                    else:
                        self.image_widget.window_funcs["t"].func = getattr(
                            np, self.proj
                        )

            # Window size for projections
            imgui.set_next_item_width(hello_imgui.em_size(15))
            winsize_changed, new_winsize = imgui.input_int(
                "Window Size", self.window_size, step=1, step_fast=2
            )
            set_tooltip(
                "Size of the temporal window (in frames) used for projection."
                " E.g. a value of 3 averages over 3 consecutive frames."
            )
            if winsize_changed and new_winsize > 0:
                self.window_size = new_winsize
                self.logger.info(f"New Window Size: {new_winsize}")

            # Gaussian Filter
            imgui.set_next_item_width(hello_imgui.em_size(15))
            gaussian_changed, new_gaussian_sigma = imgui.slider_float(
                label="sigma",
                v=self.gaussian_sigma,
                v_min=0.0,
                v_max=20.0,
            )
            set_tooltip(
                "Apply a Gaussian blur to the preview image. Sigma is in pixels; larger values yield stronger smoothing."
            )
            if gaussian_changed:
                self.gaussian_sigma = new_gaussian_sigma

            imgui.end_group()

            imgui.pop_style_var()

            # Section: Scan-phase Correction
            if not self.is_mbo_scan:
                return

            imgui.spacing()
            imgui.separator()
            imgui.text_colored(
                imgui.ImVec4(0.8, 0.8, 0.2, 1.0), "Scan-Phase Correction"
            )
            imgui.separator()
            imgui.begin_group()

            imgui.set_next_item_width(hello_imgui.em_size(10))
            phase_changed, phase_value = imgui.checkbox(
                "Fix Phase",
                self._fix_phase,
            )
            set_tooltip("Enable to apply scan-phase correction to interleaved lines.")
            if phase_changed:
                self.fix_phase = phase_value
                self.logger.info(f"Fix Phase: {phase_value}")

            imgui.columns(2, "offsets", False)
            for i, iw in enumerate(self.image_widget.data):
                ofs = self.current_offset[i]
                is_sequence = isinstance(ofs, (list, np.ndarray, tuple))

                # Compute the “maximum absolute offset” we'll use to decide if text should be red
                if is_sequence:
                    # turn it into a flat Python list of floats
                    ofs_list = [float(x) for x in ofs]
                    max_abs_offset = max(abs(x) for x in ofs_list) if ofs_list else 0.0
                else:
                    ofs_list = None
                    max_abs_offset = abs(ofs)

                imgui.text(f"{self._array_type} {i + 1}:")
                imgui.next_column()

                if is_sequence:
                    display_text = "avg."
                    if len(ofs_list) > 1:
                        display_text += f" {np.round(np.mean(ofs_list), 2)}"
                    else:
                        display_text += f" {np.round(ofs_list[0], 2):.3f}"
                else:
                    display_text = f"{np.round(ofs, 2):.3f}"

                # ffset ≥ self.max_offset, color red
                if max_abs_offset >= self.max_offset:
                    imgui.push_style_color(
                        imgui.Col_.text, imgui.ImVec4(1.0, 0.0, 0.0, 1.0)
                    )

                imgui.text(display_text)

                if max_abs_offset >= self.max_offset:
                    imgui.pop_style_color()

                # show all frame offsets in a tooltip
                if is_sequence and imgui.is_item_hovered():
                    imgui.begin_tooltip()
                    imgui.text_colored(
                        imgui.ImVec4(0.8, 0.8, 0.2, 1.0), "Per‐frame offsets:"
                    )
                    for frame_idx, val in enumerate(ofs_list):
                        imgui.text(f"  frame {frame_idx}: {val:.3f}")
                    imgui.end_tooltip()
                imgui.next_column()
            imgui.columns(1)

            imgui.set_next_item_width(hello_imgui.em_size(10))
            upsample_changed, upsample_val = imgui.input_int(
                "Upsample", self._phase_upsample, step=1, step_fast=2
            )
            set_tooltip(
                "Phase-correction upsampling factor: interpolates the image by this integer factor to improve subpixel alignment."
            )
            if upsample_changed:
                self.phase_upsample = max(1, upsample_val)
                self.logger.info(f"New upsample: {upsample_val}")

            imgui.set_next_item_width(hello_imgui.em_size(10))
            border_changed, border_val = imgui.input_int(
                "Exclude border-px", self._border, step=1, step_fast=2
            )
            set_tooltip(
                "Number of pixels to exclude from the edges of the image when computing the scan-phase offset."
            )
            if border_changed:
                self.border = max(0, border_val)
                self.logger.info(f"New border: {border_val}")

            imgui.set_next_item_width(hello_imgui.em_size(10))
            max_offset_changed, max_offset = imgui.input_int(
                "max-offset", self._max_offset, step=1, step_fast=2
            )
            set_tooltip(
                "Maximum allowed pixel shift (in pixels) when estimating the scan-phase offset."
            )
            if max_offset_changed:
                self.max_offset = max(1, max_offset)
                self.logger.info(f"New max-offset: {max_offset}")

            imgui.end_group()
            imgui.separator()

        imgui.separator()
        draw_zstats_progress(self)
        draw_saveas_progress(self)

    def get_raw_frame(self) -> tuple[ndarray, ...]:
        idx = self.image_widget.current_index
        t = idx.get("t", 0)
        z = idx.get("z", 0)
        return tuple(ndim_to_frame(arr, t, z) for arr in self.image_widget.data)

    def gui_progress_callback(self, fraction, current_plane):
        """Callback for save_as progress updates."""
        self._saveas_current_index = current_plane
        self._saveas_progress = fraction
        self._saveas_done = fraction >= 1.0

    def update_frame_apply(self):
        """Update the frame_apply function of the image widget."""
        self.image_widget.frame_apply = {
            i: partial(self._combined_frame_apply, arr_idx=i)
            for i in range(len(self.image_widget.managed_graphics))
        }

    def _combined_frame_apply(self, frame: np.ndarray, arr_idx: int = 0) -> np.ndarray:
        """alter final frame only once, in ImageWidget.frame_apply"""
        if self._gaussian_sigma > 0:
            frame = gaussian_filter(frame, sigma=self.gaussian_sigma)
        if (not self.is_mbo_scan) and self._fix_phase:
            frame = apply_scan_phase_offsets(frame, self.current_offset[arr_idx])
        if self.proj == "mean-sub" and self._zstats_means:
            z = self.image_widget.current_index.get("z", 0)
            frame = frame - self._zstats_means[arr_idx][z]
        return frame

    def _compute_zstats_single_roi(self, roi, fpath):
        arr = imread(fpath)
        arr.roi = roi

        if HAS_TORCH and isinstance(arr, torch.Tensor):
            arr = arr[:]  # make sure it's dense

        if arr.ndim == 3:
            arr = arr[:, np.newaxis, :, :]

        stats = {"mean": [], "std": [], "snr": []}
        means = []

        self._tiff_lock = threading.Lock()
        for z in range(self.nz):
            self.logger.info(
                f"--- Processing Z-plane {z + 1}/{self.nz} for ROI {roi + 1} --"
            )
            with self._tiff_lock:
                stack = arr[:, z]
                if hasattr(stack, "astype"):  # works for Dask and NumPy
                    stack = stack.astype(np.float32)
                if hasattr(stack, "compute"):
                    stack = stack.compute()

                mean_img = np.mean(stack, axis=0)
                std_img = np.std(stack, axis=0)
                snr_img = np.divide(mean_img, std_img + 1e-5, where=(std_img > 1e-5))

                stats["mean"].append(float(np.mean(mean_img)))
                stats["std"].append(float(np.mean(std_img)))
                stats["snr"].append(float(np.mean(snr_img)))
                means.append(mean_img)

                self.logger.info(
                    f"ROI {roi + 1} - Z-plane {z + 1}: "
                    f"Mean: {stats['mean'][-1]:.2f}, "
                    f"Std: {stats['std'][-1]:.2f}, "
                    f"SNR: {stats['snr'][-1]:.2f}",
                )

                self._zstats_progress[roi] = (z + 1) / self.nz
                self._zstats_current_z[roi] = z

        self._zstats[roi] = stats
        self._zstats_means[roi] = np.stack(means)
        self._zstats_done[roi] = True

    def compute_zstats(self):
        if not self.image_widget or not self.image_widget.data:
            return
        if self.rois is not None:
            for roi in range(self.rois):
                threading.Thread(
                    target=self._compute_zstats_single_roi, args=(roi, self.fpath), daemon=True
                ).start()
        else:
            threading.Thread(
                target=self._compute_zstats_single_roi, args=(self.rois, self.fpath), daemon=True
            ).start()

        # for data_ix, arr in enumerate(self.image_widget.data):
        #     self.logger.debug(f"Sending array index {data_ix} for z-stat computation..")
        #     threading.Thread(
        #         target=self._compute_zstats_single_roi, args=(data_ix, arr), daemon=True
        #     ).start()



# force import to ensure glfw is initialized before fastplotlib
import glfw
from rendercanvas.glfw import GlfwRenderCanvas, loop

from .run_gui import run_gui
from .imgui import PreviewDataWidget

__all__ = [
    "PreviewDataWidget",
    "run_gui",
]



import logging
import time
from imgui_bundle import imgui
from .. import log

GUI_LOGGERS = log.get_package_loggers()


class GuiLogHandler(logging.Handler):
    def __init__(self, gui_logger):
        super().__init__()
        self.gui_logger = gui_logger

    def emit(self, record):
        t = time.strftime("%H:%M:%S")
        level_str = {
            logging.DEBUG: "debug",
            logging.INFO: "info",
            logging.WARNING: "warning",
            logging.ERROR: "error",
            logging.CRITICAL: "error",
        }.get(record.levelno, "info")
        # extract only the last segment of the logger name, e.g. "mbo.scan" → "scan"
        name = record.name.split(".")[-1]
        self.gui_logger.messages.append((t, level_str, name, self.format(record)))


class GuiLogger:
    def __init__(self):
        self.show = True
        self.filters = {"debug": True, "info": True, "warning": True}
        self.messages = []  # now holds tuples (time, level, logger_name, text)
        self.window_flags = imgui.WindowFlags_.none
        self.active_loggers = {name: True for name in GUI_LOGGERS}
        # self.sub_loggers = log.get_all_loggers()

    def draw(self):
        # Log‐level checkboxes
        _, self.filters["debug"] = imgui.checkbox("Debug", self.filters["debug"])
        imgui.same_line()
        _, self.filters["info"] = imgui.checkbox("Info", self.filters["info"])
        imgui.same_line()
        _, self.filters["warning"] = imgui.checkbox("warning", self.filters["warning"])

        imgui.separator()

        # Sub‐logger toggles
        for name in list(self.active_loggers):
            imgui.push_id(f"logger_{name}")
            changed, state = imgui.checkbox(
                f"Logger: {name}", self.active_loggers[name]
            )
            if changed:
                self.active_loggers[name] = state
                if state:
                    log.enable(name)
                else:
                    log.disable(name)
            imgui.pop_id()

        imgui.separator()
        imgui.begin_child("##debug_scroll", imgui.ImVec2(0, 0), False)

        # iterate newest‐first
        for t, lvl, logger_name, m in reversed(self.messages):
            if not self.filters.get(lvl, False):
                continue
            if not self.active_loggers.get(logger_name, False):
                continue

            # color by severity
            if lvl == "debug":
                col = imgui.ImVec4(0.6, 0.6, 0.6, 1)
            elif lvl == "info":
                col = imgui.ImVec4(1.0, 1.0, 1.0, 1)
            else:  # "warn" or "error"
                col = imgui.ImVec4(1.0, 0.3, 0.3, 1)
            imgui.text_colored(col, f"[{t}] [{logger_name}] {m}")
        imgui.end_child()

    def draw2(self):
        # Log‐level checkboxes
        _, self.filters["debug"] = imgui.checkbox("Debug", self.filters["debug"])
        imgui.same_line()
        _, self.filters["info"] = imgui.checkbox("Info", self.filters["info"])
        imgui.same_line()
        _, self.filters["warning"] = imgui.checkbox("warning", self.filters["warning"])

        imgui.separator()

        # Sub‐logger toggles
        for name in list(self.active_loggers):
            imgui.push_id(f"logger_{name}")
            changed, state = imgui.checkbox(
                f"Logger: {name}", self.active_loggers[name]
            )
            if changed:
                self.active_loggers[name] = state
                if state:
                    log.enable(name)
                else:
                    log.disable(name)
            imgui.pop_id()

        imgui.separator()
        imgui.begin_child("##debug_scroll", imgui.ImVec2(0, 0), False)
        for t, lvl, logger_name, m in self.messages:
            # skip if this level is unchecked or this logger is disabled
            if not self.filters.get(lvl, False):
                continue
            if not self.active_loggers.get(logger_name, False):
                continue
            col = {
                "debug": imgui.ImVec4(0.8, 0.8, 0.8, 1),
                "info": imgui.ImVec4(1.0, 1.0, 1.0, 1),
                "warning": imgui.ImVec4(1.0, 0.3, 0.3, 1),
            }[lvl]
            imgui.text_colored(col, f"[{t}] [{logger_name}] {m}")
        imgui.end_child()



import threading
import time
from dataclasses import dataclass
from pathlib import Path

import numpy as np

from imgui_bundle import imgui, imgui_ctx, portable_file_dialogs as pfd
from mbo_utilities import get_metadata, get_mbo_dirs
from mbo_utilities._parsing import _make_json_serializable
from mbo_utilities.graphics._widgets import set_tooltip
from mbo_utilities.lazy_array import imread

try:
    import lbm_suite2p_python as lsp

    HAS_LSP = True
except ImportError:
    HAS_LSP = False
    lsp = None

try:
    import masknmf

    HAS_MASKNMF = True
except ImportError:
    HAS_MASKNMF = False
    masknmf = None


USER_PIPELINES = ["suite2p", "masknmf"]


@dataclass
class Suite2pSettings:
    do_registration: bool = True
    align_by_chan: int = 1
    nimg_init: int = 300
    batch_size: int = 500
    maxregshift: float = 0.1
    smooth_sigma: float = 1.15
    smooth_sigma_time: float = 0.0
    keep_movie_raw: bool = False
    two_step: bool = False
    reg_tif: bool = False
    reg_tif_chan2: bool = False
    subpixel: int = 10
    th_badframes: float = 1.0
    norm_frames: bool = True
    force_refimg: bool = False
    pad_fft: bool = False

    soma_crop: bool = True
    use_builtin_classifier: bool = False
    classifier_path: str = ""

    roidetect: bool = True
    sparse_mode: bool = True
    spatial_scale: int = 0
    connected: bool = True
    threshold_scaling: float = 1.0
    spatial_hp_detect: int = 25
    max_overlap: float = 0.75
    high_pass: int = 100
    smooth_masks: bool = True
    max_iterations: int = 20
    nbinned: int = 5000
    denoise: bool = False

    def to_dict(self):
        return {
            field.name: getattr(self, field.name)
            for field in self.__dataclass_fields__.values()  # type: ignore # noqa
        }

    def to_file(self, filepath):
        """Save settings to a JSON file."""
        np.save(filepath, self.to_dict(), allow_pickle=True)


def draw_tab_process(self):
    """Draws the pipeline selection and configuration section."""

    if not hasattr(self, "_rectangle_selectors"):
        self._rectangle_selectors = {}
    if not hasattr(self, "_current_pipeline"):
        self._current_pipeline = USER_PIPELINES[0]
    if not hasattr(self, "_install_error"):
        self._install_error = False
    if not hasattr(self, "_show_red_text"):
        self._show_red_text = False
    if not hasattr(self, "_show_green_text"):
        self._show_green_text = False
    if not hasattr(self, "_show_install_button"):
        self._show_install_button = False
    if not hasattr(self, "_region_idx"):
        self._region_idx = 0
    if not hasattr(self, "_subregions"):
        self._subregions = {}
    if not hasattr(self, "_array_type"):
        self._array_type = "array"

    imgui.begin_group()
    imgui.dummy(imgui.ImVec2(0, 5))

    imgui.text_colored(
        imgui.ImVec4(0.8, 1.0, 0.2, 1.0), "Spatial-crop before processing:"
    )

    for i, graphic in enumerate(self.image_widget.managed_graphics):
        selected = self._rectangle_selectors.get(i) is not None
        label = f"{'Remove Crop Selector: ' if selected else 'Add Crop Selector: '}{self._array_type} {i + 1}"
        if imgui.button(label):
            g = self.image_widget.managed_graphics[i]
            sel = self._rectangle_selectors.get(i)
            if sel:  # already exists → remove
                self.image_widget.figure[0, i].delete_graphic(sel)
                self._rectangle_selectors[i] = None
            else:  # doesn’t exist → add
                g.add_rectangle_selector()
                self._rectangle_selectors[i] = g._plot_area.selectors[0]

    imgui.dummy(imgui.ImVec2(0, 5))
    imgui.separator()

    imgui.text_colored(
        imgui.ImVec4(0.8, 1.0, 0.2, 1.0), "Select a processing pipeline:"
    )

    current_display_idx = USER_PIPELINES.index(self._current_pipeline)
    changed, selected_idx = imgui.combo("Pipeline", current_display_idx, USER_PIPELINES)

    if changed:
        self._current_pipeline = USER_PIPELINES[selected_idx]
    set_tooltip("Select a processing pipeline to configure.")

    if self._current_pipeline == "suite2p":
        draw_section_suite2p(self)
    elif self._current_pipeline == "masknmf":
        draw_section_masknmf(self)
    imgui.spacing()
    imgui.end_group()


def draw_section_suite2p(self):
    imgui.spacing()
    with imgui_ctx.begin_child("##Processing"):
        avail_w = imgui.get_content_region_avail().x * 0.3
        imgui.push_item_width(avail_w)
        imgui.new_line()
        imgui.separator_text("Registration Settings")
        _, self.s2p.do_registration = imgui.checkbox(
            "Do Registration", self.s2p.do_registration
        )
        set_tooltip("Whether or not to run registration")
        _, self.s2p.nimg_init = imgui.input_int("Initial Frames", self.s2p.nimg_init)
        set_tooltip(
            "Shave off this many frames from the start of the movie for registration"
        )
        _, self.s2p.batch_size = imgui.input_int("Batch Size", self.s2p.batch_size)
        set_tooltip("Number of frames to process in each batch during registration")
        _, self.s2p.maxregshift = imgui.input_float(
            "Max Shift Fraction", self.s2p.maxregshift
        )
        set_tooltip("Maximum allowed shift as a fraction of the image size")
        _, self.s2p.smooth_sigma = imgui.input_float(
            "Smooth Sigma", self.s2p.smooth_sigma
        )
        set_tooltip(
            "Sigma for Gaussian smoothing of the image. Keep it low (less than 1.0) for high-frequency data."
        )
        _, self.s2p.smooth_sigma_time = imgui.input_float(
            "Smooth Sigma Time", self.s2p.smooth_sigma_time
        )
        set_tooltip("Sigma for temporal smoothing of the image.")
        _, self.s2p.keep_movie_raw = imgui.checkbox(
            "Keep Raw Movie", self.s2p.keep_movie_raw
        )
        set_tooltip(
            "Whether to keep the raw movie data after processing, kept as a .bin file."
        )
        _, self.s2p.two_step = imgui.checkbox(
            "Two-Step Registration", self.s2p.two_step
        )
        set_tooltip(
            "Register once to a reference image, then register to the mean of the registered movie."
        )
        _, self.s2p.reg_tif = imgui.checkbox("Export Registered TIFF", self.s2p.reg_tif)
        set_tooltip(
            "Export the registered movie as a TIFF file. Saved in save/path/reg_tiff/reg_tif_chan1_000N.tif"
        )
        _, self.s2p.subpixel = imgui.input_int("Subpixel Precision", self.s2p.subpixel)
        set_tooltip(
            "Subpixel precision for registration. Higher values may improve accuracy but increase processing time."
        )
        _, self.s2p.th_badframes = imgui.input_float(
            "Bad Frame Threshold", self.s2p.th_badframes
        )
        set_tooltip(
            "Threshold for detecting bad frames during registration. Frames with a value above this threshold will be considered bad."
        )
        _, self.s2p.norm_frames = imgui.checkbox(
            "Normalize Frames", self.s2p.norm_frames
        )
        set_tooltip(
            "Whether to normalize frames during registration. This can help with illumination variations."
        )
        _, self.s2p.force_refimg = imgui.checkbox(
            "Use Stored refImg", self.s2p.force_refimg
        )
        set_tooltip(
            "Use a stored reference image for registration instead of computing one from the movie."
        )
        _, self.s2p.pad_fft = imgui.checkbox("Pad FFT Image", self.s2p.pad_fft)
        set_tooltip(
            "Whether to pad the FFT image during registration. This can help with aliasing effects."
        )

        imgui.spacing()
        imgui.separator_text("Classification Settings")
        _, self.s2p.soma_crop = imgui.checkbox("Soma Crop", self.s2p.soma_crop)
        set_tooltip("Crop the movie to the soma region before processing.")
        _, self.s2p.use_builtin_classifier = imgui.checkbox(
            "Use Builtin Classifier", self.s2p.use_builtin_classifier
        )
        set_tooltip(
            "Use the built-in classifier for detecting ROIs. If unchecked, a custom classifier path is required."
        )
        _, self.s2p.classifier_path = imgui.input_text(
            "Classifier Path", self.s2p.classifier_path, 256
        )
        set_tooltip(
            "Path to a custom classifier file. If the built-in classifier is not used, this path must be provided."
        )

        imgui.separator_text("ROI Detection Settings")
        _, self.s2p.roidetect = imgui.checkbox("Detect ROIs", self.s2p.roidetect)
        set_tooltip("Whether to detect ROIs in the movie.")
        _, self.s2p.sparse_mode = imgui.checkbox("Sparse Mode", self.s2p.sparse_mode)
        set_tooltip("Sparse_mode=True is recommended for soma, False for dendrites.")
        _, self.s2p.spatial_scale = imgui.input_int(
            "Spatial Scale", self.s2p.spatial_scale
        )
        set_tooltip(
            "what the optimal scale of the recording is in pixels. if set to 0, then the algorithm determines it automatically (recommend this on the first try). If it seems off, set it yourself to the following values: 1 (=6 pixels), 2 (=12 pixels), 3 (=24 pixels), or 4 (=48 pixels)."
        )
        _, self.s2p.connected = imgui.checkbox("Connected ROIs", self.s2p.connected)
        set_tooltip(
            "Whether to use connected components for ROI detection. If False, ROIs will be detected independently."
        )
        _, self.s2p.threshold_scaling = imgui.input_float(
            "Threshold Scaling", self.s2p.threshold_scaling
        )
        set_tooltip(
            "Scaling factor for the threshold used in ROI detection. Generally (NOT always), higher values will result in fewer ROIs being detected."
        )
        _, self.s2p.spatial_hp_detect = imgui.input_int(
            "Spatial HP Filter", self.s2p.spatial_hp_detect
        )
        set_tooltip(
            "Spatial high-pass filter size for ROI detection. A value of 25 is recommended for most datasets."
        )
        _, self.s2p.max_overlap = imgui.input_float("Max Overlap", self.s2p.max_overlap)
        set_tooltip(
            "Maximum allowed overlap between detected ROIs. If two ROIs overlap more than this value, the one with the lower signal will be discarded."
        )
        _, self.s2p.high_pass = imgui.input_int("High Pass", self.s2p.high_pass)
        set_tooltip(
            "High-pass filter size for ROI detection. A value of 100 is recommended for most datasets."
        )
        _, self.s2p.smooth_masks = imgui.checkbox("Smooth Masks", self.s2p.smooth_masks)
        set_tooltip(
            "Whether to smooth the detected masks. This can help with noise but may also remove small ROIs."
        )
        _, self.s2p.max_iterations = imgui.input_int(
            "Max Iterations", self.s2p.max_iterations
        )
        set_tooltip(
            "Maximum number of iterations for the ROI detection algorithm. More than 100 is likely redundant."
        )
        _, self.s2p.nbinned = imgui.input_int("Max Binned Frames", self.s2p.nbinned)
        set_tooltip(
            "Maximum number of frames to bin for ROI detection. More than 5000 is likely redundant."
        )
        _, self.s2p.denoise = imgui.checkbox("Denoise Movie", self.s2p.denoise)
        set_tooltip(
            "Whether to denoise the movie before processing. This can help with noise but may also remove small ROIs."
        )

        imgui.spacing()
        imgui.input_text("Save folder", self._saveas_outdir, 256)
        imgui.same_line()
        if imgui.button("Browse"):
            home = Path().home()
            res = pfd.select_folder(str(home))
            if res:
                self._saveas_outdir = res.result()

        imgui.separator()
        if imgui.button("Run"):
            self.logger.info("Running Suite2p pipeline...")
            run_process(self)
            self.logger.info("Suite2p pipeline completed.")
        if self._install_error:
            imgui.same_line()
            if self._show_red_text:
                imgui.text_colored(
                    imgui.ImVec4(1.0, 0.0, 0.0, 1.0),
                    "Error: lbm_suite2p_python is not installed.",
                )
            if self._show_green_text:
                imgui.text_colored(
                    imgui.ImVec4(1.0, 0.0, 0.0, 1.0),
                    "lbm_suite2p_python install success.",
                )
            if self._show_install_button:
                if imgui.button("Install"):
                    import subprocess

                    self.logger.log("info", "Installing lbm_suite2p_python...")
                    try:
                        subprocess.check_call(["pip", "install", "lbm_suite2p_python"])
                        self.logger.log("info", "Installation complete.")
                        self._install_error = False
                        self._show_red_text = False
                        self._show_green_text = True
                    except Exception as e:
                        try:
                            self.logger.log(
                                "error",
                                f"Installation failed: {e}",
                            )
                            subprocess.check_call(
                                ["uv", "pip", "install", "lbm_suite2p_python"]
                            )
                            self._show_red_text = False
                            self._show_green_text = True
                        except Exception as e:
                            self.logger.log("error", f"Installation failed: {e}")
                            self._show_red_text = True
                            self._show_install_button = False
                            self._show_green_text = True

        imgui.pop_item_width()


def draw_section_masknmf(self):
    imgui.text("MaskNMF is not yet implemented in this version.")


def run_process(self):
    """Runs the selected processing pipeline."""
    if self._current_pipeline == "suite2p":
        self.logger.info(f"Running Suite2p pipeline with settings: {self.s2p}")
        try:
            import lbm_suite2p_python as lsp
        except ImportError:
            self.logger.warning(
                "error",
                "lbm_suite2p_python is not installed. Please install it to run the Suite2p pipeline.",
            )
            self._install_error = True
            return
    if not self._install_error:
        for i, arr in enumerate(self.image_widget.data):
            kwargs = {"self": self, "arr_idx": i}
            threading.Thread(target=run_plane_from_data, kwargs=kwargs).start()
    elif self._current_pipeline == "masknmf":
        self.logger.info("Running MaskNMF pipeline (not yet implemented).")
    else:
        self.logger.error(f"Unknown pipeline selected: {self._current_pipeline}")


def run_plane_from_data(self, arr_idx):
    if not HAS_LSP:
        self.logger.error(
            "lbm_suite2p_python is not installed. Please install it to run the Suite2p pipeline."
        )
        self._install_error = True
        return

    data_shape = self.image_widget.data[arr_idx].shape
    metadata = getattr(self, "metadata", None)
    input_file = None

    dims = self.image_widget.current_index
    if "z" in dims:
        current_z = self.image_widget.current_index["z"]
    else:
        current_z = 0

    if arr_idx in self._rectangle_selectors.keys():
        ind_xy = self._rectangle_selectors[arr_idx].get_selected_indices()
        ind_x = list(ind_xy[0])
        ind_y = list(ind_xy[1])
        self.logger.debug(f"Sub-indices selected: {ind_xy}")
    else:
        ind_x, ind_y = slice(None), slice(None)

    if self.is_mbo_scan:
        # TODO
        raise NotImplementedError()
        # save_as(
        #     self.image_widget.data[arr_idx],
        #     self._saveas_outdir,
        #     planes=[current_z + 1],
        #     ext=".bin",
        # )

    # move to property?
    if not self._saveas_outdir:
        current_time_fmt = time.strftime("%Y%m%d_%H%M%S")
        self._saveas_outdir = get_mbo_dirs()["data"].joinpath(
            f"{current_time_fmt}_{self._current_pipeline}_output"
        )

    if len(data_shape) == 4:
        data = self.image_widget.data[arr_idx][:, current_z, ind_x, ind_y]
    elif len(data_shape) == 3:
        data = self.image_widget.data[arr_idx][:, ind_x, ind_y]
    else:
        data = self.image_widget.data[arr_idx][ind_x, ind_y]

    self.logger.info(f"shape of selected data: {data.shape}")
    out_dir = Path(self._saveas_outdir)
    out_dir.mkdir(exist_ok=True)

    self.fpath = self.fpath
    loader = imread(self.fpath)
    if hasattr(loader, "roi") or hasattr(loader, "rois"):
        self.logger.info("Using LazyArrayLoader with ROI support. ")
        arr = loader.rois

    if self.fpath.suffix.lower() in [".tif", ".tiff"]:
        metadata = get_metadata(self.fpath)
        metadata = _make_json_serializable(metadata)
        input_file = out_dir / "plane_7.tif"
        metadata["shape"] = data.shape
        # TODO
        raise NotImplementedError()
        # save_as(
        #     data,
        #     input_file,
        #     metadata
        # )
        # tifffile.imwrite(input_file, data, metadata=metadata)
        # self.logger.info(f"Temporary file created:"
        #             f" {input_file}")

    elif self.fpath.suffix.lower() == ".bin":
        metadata = np.load(self.fpath.parent / "ops.npy", allow_pickle=True).item()
        input_file = self.fpath

    self.logger.info(f"Metadata provided:")
    for k, v in metadata.items():
        self.logger.info(f"{k}: {v}")
    ops = self.s2p.to_dict()
    self.logger.info(f"User ops provided:")
    for k, v in ops.items():
        self.logger.info(f"{k}: {v}")
    ops.update(metadata)
    if input_file:
        lsp.run_plane(input_file, out_dir, ops=ops)
        self.logger.info(f"Plane 7 saved to {out_dir / 'plane_7.tif'}")
    else:
        self.logger.error("No valid input file found for processing.")
        return



from __future__ import annotations

import re
import json
import os
import struct
from pathlib import Path
from typing import Any

import numpy as np
import tifffile
from tifffile import read_scanimage_metadata, matlabstr2py
from tifffile.tifffile import bytes2str, read_json, FileHandle


def _params_from_metadata_caiman(metadata):
    """
    Generate parameters for CNMF from metadata.

    Based on the pixel resolution and frame rate, the parameters are set to reasonable values.

    Parameters
    ----------
    metadata : dict
        Metadata dictionary resulting from `lcp.get_metadata()`.

    Returns
    -------
    dict
        Dictionary of parameters for lbm_mc.

    """
    params = _default_params_caiman()

    if metadata is None:
        print("No metadata found. Using default parameters.")
        return params

    params["main"]["fr"] = metadata["frame_rate"]
    params["main"]["dxy"] = metadata["pixel_resolution"]

    # typical neuron ~16 microns
    gSig = round(16 / metadata["pixel_resolution"][0]) / 2
    params["main"]["gSig"] = (int(gSig), int(gSig))

    gSiz = (4 * gSig + 1, 4 * gSig + 1)
    params["main"]["gSiz"] = gSiz

    max_shifts = [int(round(10 / px)) for px in metadata["pixel_resolution"]]
    params["main"]["max_shifts"] = max_shifts

    strides = [int(round(64 / px)) for px in metadata["pixel_resolution"]]
    params["main"]["strides"] = strides

    # overlap should be ~neuron diameter
    overlaps = [int(round(gSig / px)) for px in metadata["pixel_resolution"]]
    if overlaps[0] < gSig:
        print("Overlaps too small. Increasing to neuron diameter.")
        overlaps = [int(gSig)] * 2
    params["main"]["overlaps"] = overlaps

    rf_0 = (strides[0] + overlaps[0]) // 2
    rf_1 = (strides[1] + overlaps[1]) // 2
    rf = int(np.mean([rf_0, rf_1]))

    stride = int(np.mean([overlaps[0], overlaps[1]]))

    params["main"]["rf"] = rf
    params["main"]["stride"] = stride

    return params


def _default_params_caiman():
    """
    Default parameters for both registration and CNMF.
    The exception is gSiz being set relative to gSig.

    Returns
    -------
    dict
        Dictionary of default parameter values for registration and segmentation.

    Notes
    -----
    This will likely change as CaImAn is updated.
    """
    gSig = 6
    gSiz = (4 * gSig + 1, 4 * gSig + 1)
    return {
        "main": {
            # Motion correction parameters
            "pw_rigid": True,
            "max_shifts": [6, 6],
            "strides": [64, 64],
            "overlaps": [8, 8],
            "min_mov": None,
            "gSig_filt": [0, 0],
            "max_deviation_rigid": 3,
            "border_nan": "copy",
            "splits_els": 14,
            "upsample_factor_grid": 4,
            "use_cuda": False,
            "num_frames_split": 50,
            "niter_rig": 1,
            "is3D": False,
            "splits_rig": 14,
            "num_splits_to_process_rig": None,
            # CNMF parameters
            "fr": 10,
            "dxy": (1.0, 1.0),
            "decay_time": 0.4,
            "p": 2,
            "nb": 3,
            "K": 20,
            "rf": 64,
            "stride": [8, 8],
            "gSig": gSig,
            "gSiz": gSiz,
            "method_init": "greedy_roi",
            "rolling_sum": True,
            "use_cnn": False,
            "ssub": 1,
            "tsub": 1,
            "merge_thr": 0.7,
            "bas_nonneg": True,
            "min_SNR": 1.4,
            "rval_thr": 0.8,
        },
        "refit": True,
    }


def _params_from_metadata_suite2p(metadata, ops):
    """
    Tau is 0.7 for GCaMP6f, 1.0 for GCaMP6m, 1.25-1.5 for GCaMP6s
    """
    if metadata is None:
        print("No metadata found. Using default parameters.")
        return ops

    # typical neuron ~16 microns
    ops["fs"] = metadata["frame_rate"]
    ops["nplanes"] = 1
    ops["nchannels"] = 1
    ops["do_bidiphase"] = 0
    ops["do_regmetrics"] = True

    # suite2p iterates each plane and takes ops['dxy'][i] where i is the plane index
    ops["dx"] = [metadata["pixel_resolution"][0]]
    ops["dy"] = [metadata["pixel_resolution"][1]]

    return ops


def report_missing_metadata(file: os.PathLike | str):
    tiff_file = tifffile.TiffFile(file)
    if not tiff_file.software == "SI":
        print(f"Missing SI software tag.")
    if not tiff_file.description[:6] == "state.":
        print(f"Missing 'state' software tag.")
    if not "scanimage.SI" in tiff_file.description[-256:]:
        print(f"Missing 'scanimage.SI' in description tag.")


def has_mbo_metadata(file: os.PathLike | str) -> bool:
    """
    Check if a TIFF file has metadata from the Miller Brain Observatory.

    Specifically, this checks for tiff_file.shaped_metadata, which is used to store system and user
    supplied metadata.

    Parameters
    ----------
    file: os.PathLike
        Path to the TIFF file.

    Returns
    -------
    bool
        True if the TIFF file has MBO metadata; False otherwise.
    """
    if not file or not isinstance(file, (str, os.PathLike)):
        raise ValueError(
            "Invalid file path provided: must be a string or os.PathLike object."
            f"Got: {file} of type {type(file)}"
        )
    # Tiffs
    if Path(file).suffix in [".tif", ".tiff"]:
        try:
            tiff_file = tifffile.TiffFile(file)
            if (
                hasattr(tiff_file, "shaped_metadata")
                and tiff_file.shaped_metadata is not None
            ):
                return True
            else:
                return False
        except Exception:
            return False
    return False


def is_raw_scanimage(file: os.PathLike | str) -> bool:
    """
    Check if a TIFF file is a raw ScanImage TIFF.

    Parameters
    ----------
    file: os.PathLike
        Path to the TIFF file.

    Returns
    -------
    bool
        True if the TIFF file is a raw ScanImage TIFF; False otherwise.
    """
    if not file or not isinstance(file, (str, os.PathLike)):
        return False
    elif Path(file).suffix not in [".tif", ".tiff"]:
        return False
    try:
        tiff_file = tifffile.TiffFile(file)
        if (
            # TiffFile.shaped_metadata is where we store metadata for processed tifs
            # if this is not empty, we have a processed file
            # otherwise, we have a raw scanimage tiff
            hasattr(tiff_file, "shaped_metadata")
            and tiff_file.shaped_metadata is not None
            and isinstance(tiff_file.shaped_metadata, (list, tuple))
        ):
            return False
        else:
            if tiff_file.scanimage_metadata is None:
                print(f"No ScanImage metadata found in {file}.")
                return False
            return True
    except Exception:
        return False


def get_metadata(file: os.PathLike | str, z_step=None, verbose=False, strict=False):
    """
    Extract metadata from a TIFF file produced by ScanImage or processed via the save_as function.

    This function opens the given TIFF file and retrieves critical imaging parameters and acquisition details.
    It supports both raw ScanImage TIFFs and those modified by downstream processing. If the file contains
    raw ScanImage metadata, the function extracts key fields such as channel information, number of frames,
    field-of-view, pixel resolution, and ROI details. When verbose output is enabled, the complete metadata
    document is returned in addition to the parsed key values.

    Parameters
    ----------
    file : os.PathLike or str
        The full path to the TIFF file from which metadata is to be extracted.
    verbose : bool, optional
        If True, returns an extended metadata dictionary that includes all available ScanImage attributes.
        Default is False.
    z_step : float, optional
        The z-step size in microns. If provided, it will be included in the returned metadata.

    Returns
    -------
    dict
        A dictionary containing the extracted metadata (e.g., number of planes, frame rate, field-of-view,
        pixel resolution). When verbose is True, the dictionary also includes a key "all" with the full metadata
        from the TIFF header.

    Raises
    ------
    ValueError
        If no recognizable metadata is found in the TIFF file (e.g., the file is not a valid ScanImage TIFF).

    Notes
    -----
    - num_frames represents the number of frames per z-plane

    Examples
    --------
    >>> meta = get_metadata("path/to/rawscan_00001.tif")
    >>> print(meta["num_frames"])
    5345
    >>> meta = get_metadata("path/to/assembled_data.tif")
    >>> print(meta["shape"])
    (14, 5345, 477, 477)
    >>> meta_verbose = get_metadata("path/to/scanimage_file.tif", verbose=True)
    >>> print(meta_verbose["all"])
    {... Includes all ScanImage FrameData ...}
    """
    if isinstance(file, list):
        return get_metadata_batch(file)

    tiff_file = tifffile.TiffFile(file)
    # previously processed files
    if not is_raw_scanimage(file):
        return tiff_file.shaped_metadata[0]
    elif hasattr(tiff_file, "scanimage_metadata"):
        meta = tiff_file.scanimage_metadata
        if meta is None:
            return None

        si = meta.get("FrameData", {})
        if not si:
            print(f"No FrameData found in {file}.")
            return None
        series = tiff_file.series[0]
        pages = tiff_file.pages
        print("Raw tiff fully read.")

        # Extract ROI and imaging metadata
        roi_group = meta["RoiGroups"]["imagingRoiGroup"]["rois"]

        if isinstance(roi_group, dict):
            num_rois = 1
            roi_group = [roi_group]
        else:
            num_rois = len(roi_group)

        num_planes = len(si["SI.hChannels.channelSave"])

        if num_rois > 1:
            try:
                sizes = [
                    roi_group[i]["scanfields"][i]["sizeXY"] for i in range(num_rois)
                ]
                num_pixel_xys = [
                    roi_group[i]["scanfields"][i]["pixelResolutionXY"]
                    for i in range(num_rois)
                ]
            except KeyError:
                sizes = [roi_group[i]["scanfields"]["sizeXY"] for i in range(num_rois)]
                num_pixel_xys = [
                    roi_group[i]["scanfields"]["pixelResolutionXY"]
                    for i in range(num_rois)
                ]

            # see if each item in sizes is the same
            if strict:
                assert all([sizes[0] == size for size in sizes]), (
                    "ROIs have different sizes"
                )
                assert all(
                    [num_pixel_xys[0] == num_pixel_xy for num_pixel_xy in num_pixel_xys]
                ), "ROIs have different pixel resolutions"
            size_xy = sizes[0]
            num_pixel_xy = num_pixel_xys[0]
        else:
            size_xy = [roi_group[0]["scanfields"]["sizeXY"]][0]
            num_pixel_xy = [roi_group[0]["scanfields"]["pixelResolutionXY"]][0]

        # TIFF header-derived metadata
        objective_resolution = si["SI.objectiveResolution"]
        frame_rate = si["SI.hRoiManager.scanFrameRate"]

        # Field-of-view calculations
        # TODO: We may want an FOV measure that takes into account contiguous ROIs
        # As of now, this is for a single ROI
        fov_x_um = round(objective_resolution * size_xy[0])  # in microns
        fov_y_um = round(objective_resolution * size_xy[1])  # in microns
        fov_roi_um = (fov_x_um, fov_y_um)  # in microns

        pixel_resolution = (fov_x_um / num_pixel_xy[0], fov_y_um / num_pixel_xy[1])
        metadata = {
            "num_planes": num_planes,
            "fov": fov_roi_um,  # in microns
            "fov_px": tuple(num_pixel_xy),
            "num_rois": num_rois,
            "frame_rate": frame_rate,
            "pixel_resolution": np.round(pixel_resolution, 2),
            "ndim": series.ndim,
            "dtype": "int16",
            "size": series.size,
            "tiff_pages": len(pages),
            "roi_width_px": num_pixel_xy[0],
            "roi_height_px": num_pixel_xy[1],
            "objective_resolution": objective_resolution,
        }
        if verbose:
            metadata["all"] = meta
            return metadata
        else:
            return metadata
    else:
        raise ValueError(f"No metadata found in {file}.")


def get_metadata_batch(files: list[os.PathLike | str], z_step=None, verbose=False):
    """
    Extract and aggregate metadata from a list of TIFF files produced by ScanImage.

    Parameters
    ----------
    files : list of str or PathLike
        List of paths to TIFF files.
    z_step : float, optional
        Z-step in microns to include in the returned metadata.
    verbose : bool, optional
        If True, include full metadata from the first TIFF in 'all' key.

    Returns
    -------
    dict
        Aggregated metadata dictionary with total frame count and per-file page counts.
    """
    total_frames = 0
    frame_indices = []
    first_meta = None

    for i, f in enumerate(files):
        tf = tifffile.TiffFile(f)
        num_pages = len(tf.pages)
        frame_indices.append(num_pages)
        total_frames += num_pages
        if i == 0:
            if not is_raw_scanimage(f):
                base = tf.shaped_metadata[0]["image"]
            elif (
                hasattr(tf, "scanimage_metadata") and tf.scanimage_metadata is not None
            ):
                base = get_metadata(f, z_step=z_step, verbose=verbose)
            else:
                raise ValueError(f"No metadata found in {f}.")
            first_meta = base.copy()

    first_meta["num_frames"] = total_frames
    first_meta["frame_indices"] = frame_indices
    return first_meta


def params_from_metadata(metadata, base_ops, pipeline="suite2p"):
    """
    Use metadata to get sensible default pipeline parameters.

    If ops are not provided, uses suite2p.default_ops(). Sets framerate, pixel resolution, and do_metrics=True.

    Parameters
    ----------
    metadata : dict
        Result of mbo.get_metadata()
    base_ops : dict
        Ops dict to use as a base.
    pipeline : str, optional
        The pipeline to use. Default is "suite2p".
    """
    if pipeline.lower() == "caiman":
        print("Warning: CaImAn is not stable, proceed at your own risk.")
        return _params_from_metadata_caiman(metadata)
    elif pipeline.lower() == "suite2p":
        print("Setting pipeline to suite2p")
        return _params_from_metadata_suite2p(metadata, base_ops)
    else:
        raise ValueError(
            f"Pipeline {pipeline} not recognized. Use 'caiman' or 'suite2'"
        )


def read_scanimage_metadata_tifffile(
    fh: FileHandle, /
) -> tuple[dict[str, Any], dict[str, Any], int]:
    """FROM TIFFFILE for DEVELOPMENT

    Read ScanImage BigTIFF v3 or v4 static and ROI metadata from file.

    The settings can be used to read image and metadata without parsing
    the TIFF file.

    Frame data and ROI groups can alternatively be obtained from the Software
    and Artist tags of any TIFF page.

    Parameters:
        fh: Binary file handle to read from.

    Returns:
        - Non-varying frame data, parsed with :py:func:`matlabstr2py`.
        - ROI group data, parsed from JSON.
        - Version of metadata (3 or 4).

    Raises:
        ValueError: File does not contain valid ScanImage metadata.

    """
    fh.seek(0)
    try:
        byteorder, version = struct.unpack("<2sH", fh.read(4))
        if byteorder != b"II" or version != 43:
            raise ValueError("not a BigTIFF file")
        fh.seek(16)
        magic, version, size0, size1 = struct.unpack("<IIII", fh.read(16))
        if magic != 117637889 or version not in {3, 4}:
            raise ValueError(f"invalid magic {magic} or version {version} number")
    except UnicodeDecodeError as exc:
        raise ValueError("file must be opened in binary mode") from exc
    except Exception as exc:
        raise ValueError("not a ScanImage BigTIFF v3 or v4 file") from exc

    frame_data = matlabstr2py(bytes2str(fh.read(size0)[:-1]))
    roi_data = read_json(fh, "<", 0, size1, 0) if size1 > 1 else {}
    return frame_data, roi_data, version


def matlabstr(obj):
    """Convert Python dict to ScanImage-style MATLAB string."""

    def _format(v):
        if isinstance(v, list):
            if all(isinstance(i, str) for i in v):
                return "{" + " ".join(f"'{i}'" for i in v) + "}"
            return "[" + " ".join(str(i) for i in v) + "]"
        if isinstance(v, str):
            return f"'{v}'"
        if isinstance(v, bool):
            return "true" if v else "false"
        return str(v)

    return "\n".join(f"{k} = {_format(v)}" for k, v in obj.items())


def _parse_value(value_str):
    if value_str.startswith("'") and value_str.endswith("'"):
        return value_str[1:-1]
    if value_str == "true":
        return True
    if value_str == "false":
        return False
    if value_str == "NaN":
        return float("nan")
    if value_str == "Inf":
        return float("inf")
    if re.match(r"^\d+(\.\d+)?$", value_str):
        return float(value_str) if "." in value_str else int(value_str)
    if re.match(r"^\[(.*)]$", value_str):
        return [_parse_value(v.strip()) for v in value_str[1:-1].split()]
    return value_str


def _parse_key_value(parse_line):
    key_str, value_str = parse_line.split(" = ", 1)
    return key_str, _parse_value(value_str)


def parse(metadata_str):
    """
    Parses the metadata string from a ScanImage Tiff file.

    :param metadata_str:
    :return metadata_kv, metadata_json:
    """
    lines = metadata_str.split("\n")
    metadata_kv = {}
    json_portion = []
    parsing_json = False

    for line in lines:
        line = line.strip()
        if not line:
            continue
        if line.startswith("SI."):
            key, value = _parse_key_value(line)
            metadata_kv[key] = value
        elif line.startswith("{"):
            parsing_json = True
        if parsing_json:
            json_portion.append(line)
    metadata_json = json.loads("\n".join(json_portion))
    return metadata_kv, metadata_json


def find_scanimage_metadata(path):
    with tifffile.TiffFile(path) as tif:
        if hasattr(tif, "scanimage_metadata"):
            return tif.scanimage_metadata
        p = tif.pages[0]
        cand = []
        for tag in ("ImageDescription", "Software"):
            if tag in p.tags:
                cand.append(p.tags[tag].value)
        if getattr(p, "description", None):
            cand.append(p.description)
        cand.extend(str(tif.__dict__.get(k, "")) for k in tif.__dict__)
        for s in cand:
            if isinstance(s, bytes):
                s = s.decode(errors="ignore")
            m = re.search(r"{.*ScanImage.*}", s, re.S)
            if m:
                try:
                    return json.loads(m.group(0))
                except Exception:
                    return m.group(0)
    return None



import numpy as np
from pathlib import Path


def load_from_dir(plane_dir):
    """
    Load all result files generated by `masknmf.run_plane()` from a given plane directory.

    Returns
    -------
    dict
        Dictionary with the following keys:
        - 'moco': np.ndarray
              Motion-corrected data array.
        - 'pmd_demixer': object
              Deserialized PMD demixer object containing the full demixing state.
        - 'a': np.ndarray
              Spatial components array with shape (height, width, n_components).
              Each component is a 2D spatial footprint.
        - 'c': np.ndarray
              Temporal components array with shape (n_frames, n_components).
              Each column is a time series for one component.
    """
    plane_dir = Path(plane_dir).expanduser()

    results = {}
    moco_path = plane_dir / "data_reg.npy"
    pmd_demixer_path = plane_dir / "pmd_demixer.npy"
    a_path = plane_dir / "a.npy"
    c_path = plane_dir / "a.npy"

    if moco_path.exists():
        results["moco"] = np.load(moco_path, allow_pickle=True)
    if pmd_demixer_path.exists():
        results["pmd_demixer"] = np.load(pmd_demixer_path, allow_pickle=True).item()
    if a_path.exists():
        results["a"] = np.load(a_path)
    if c_path.exists():
        results["c"] = np.load(c_path)
    return results



from pathlib import Path
import os
from imgui_bundle import (
    imgui,
    imgui_md,
    hello_imgui,
    imgui_ctx,
)
from imgui_bundle import portable_file_dialogs as pfd

from mbo_utilities import get_mbo_dirs
from mbo_utilities.graphics._widgets import set_tooltip

MBO_THREADING_ENABLED = bool(
    int(os.getenv("MBO_THREADING_ENABLED", "1"))
)  # export MBO_DEV=1 to enable

MBO_ASSETS_PATH = get_mbo_dirs()["assets"]
MBO_LOGO_PATH = Path(MBO_ASSETS_PATH).joinpath("static", "logo_utilities.png")

HAS_LOGO = MBO_LOGO_PATH.exists() and MBO_LOGO_PATH.is_file()

if not HAS_LOGO:
    print("Warning: MBO logo not found at", MBO_LOGO_PATH)


class FileDialog:
    def __init__(self):
        self._assets_path = MBO_ASSETS_PATH
        self._logo_path = MBO_LOGO_PATH
        self.selected_path = None
        self._open_multi = None
        self._select_folder = None
        self._threading_enabled = MBO_THREADING_ENABLED
        self._widget_enabled = True
        self.metadata_only = False

    @property
    def widget_enabled(self):
        return self._widget_enabled

    @widget_enabled.setter
    def widget_enabled(self, value):
        self._widget_enabled = value

    @property
    def threading_enabled(self):
        return self._threading_enabled

    @threading_enabled.setter
    def threading_enabled(self, value):
        self._threading_enabled = value
        # print a warning because
        # this is generally done to fix a bug
        if not value:
            print(
                "Threading disabled,"
                " please report issues at https://github.com/MillerBrainObservatory/mbo_utilities/issues"
            )
            os.environ["MBO_THREADING_ENABLED"] = "0"
        if value:
            os.environ["MBO_THREADING_ENABLED"] = "1"

    def render(self):
        with imgui_ctx.begin_child("#fd"):
            imgui.push_id("pfd")

            # header --------------------------------------------------
            imgui.separator()
            imgui.dummy(hello_imgui.em_to_vec2(0, 0.5))

            # if HAS_LOGO:
            #     hello_imgui.image_from_asset(str(self._logo_path.expanduser().resolve().absolute()))

            imgui_md.render_unindented("""
            # General Python and shell utilities developed for the Miller Brain Observatory (MBO) workflows.

            ## Preview raw TIFFs, TIFF stacks, ScanImage files, or numpy memmaps. Load a directory of raw ScanImage files to run the data-preview widget, which allows visualization of projections, mean-subtraction, and preview scan-phase correction.

            [Docs Overview](https://millerbrainobservatory.github.io/mbo_utilities/) |
            [Assembly Guide](https://millerbrainobservatory.github.io/mbo_utilities/assembly.html) |
            [Function Examples](https://millerbrainobservatory.github.io/mbo_utilities/api/usage.html)
            """)

            imgui.dummy(hello_imgui.em_to_vec2(0, 5))

            # centre prompt ------------------------------------------
            txt = "Select a file, multiple files, or a folder to preview:"
            imgui.set_cursor_pos_x(
                (imgui.get_window_width() - imgui.calc_text_size(txt).x) * 0.5
            )
            imgui.text_colored(imgui.ImVec4(1, 0.85, 0.3, 1), txt)
            imgui.dummy(hello_imgui.em_to_vec2(0, 0.5))

            # Centered “Open File(s)” button
            bsz_file = hello_imgui.em_to_vec2(18, 2.4)
            x_file = (imgui.get_window_width() - bsz_file.x) * 0.5
            imgui.set_cursor_pos_x(x_file)
            if imgui.button("Open File(s)", bsz_file):
                self._open_multi = pfd.open_file(
                    "Select files", options=pfd.opt.multiselect
                )
            if imgui.is_item_hovered():
                imgui.set_tooltip("Open one or more TIFF / BIN files.")

            # small vertical gap
            imgui.dummy(hello_imgui.em_to_vec2(0, 1.0))

            # Centered, smaller “Select Folder” button underneath
            bsz_folder = hello_imgui.em_to_vec2(12, 2.0)
            x_folder = (imgui.get_window_width() - bsz_folder.x) * 0.5
            imgui.set_cursor_pos_x(x_folder)
            if imgui.button("Select Folder", bsz_folder):
                self._select_folder = pfd.select_folder("Select folder")
            if imgui.is_item_hovered():
                imgui.set_tooltip("Pick a directory of TIFFs.")

            # load options -------------------------------------------
            imgui.dummy(hello_imgui.em_to_vec2(0, 0.7))
            imgui.text_colored(imgui.ImVec4(1, 0.85, 0.3, 1), "Load Options")

            imgui.begin_group()
            _, self.threading_enabled = imgui.checkbox(
                "Enable Threading", self.threading_enabled
            )
            set_tooltip(
                "Enable/disable threading for the data preview widget. "
                "Useful to turn this off if you experience issues with the widget or for debugging."
                "For issues, please report here: "
                "https://github.com/MillerBrainObservatory/mbo_utilities/issues/new"
            )
            _, self.widget_enabled = imgui.checkbox(
                "Enable 'Data Preview' widget", self._widget_enabled
            )
            set_tooltip(
                "Enable/disable the 'Data Preview' widget. "
                "This widget allows you to visualize projections,"
                " mean-subtraction,"
                " and preview scan-phase correction."
            )

            _, self.metadata_only = imgui.checkbox(
                "Metadata Preview Only", self.metadata_only
            )
            set_tooltip(
                "Enable/disable the 'Data Preview' widget. "
                "This widget allows you to visualize projections,"
                " mean-subtraction,"
                " and preview scan-phase correction."
            )
            imgui.end_group()

            if self._open_multi and self._open_multi.ready():
                self.selected_path = self._open_multi.result()
                if self.selected_path:
                    hello_imgui.get_runner_params().app_shall_exit = True
                self._open_multi = None
            if self._select_folder and self._select_folder.ready():
                self.selected_path = self._select_folder.result()
                if self.selected_path:
                    hello_imgui.get_runner_params().app_shall_exit = True
                self._select_folder = None

            # quit button bottom-right -------------------------------
            qsz = hello_imgui.em_to_vec2(10, 1.8)
            imgui.set_cursor_pos(
                imgui.ImVec2(
                    imgui.get_window_width() - qsz.x - hello_imgui.em_size(1),
                    imgui.get_window_height() - qsz.y - hello_imgui.em_size(1),
                )
            )
            if imgui.button("Quit", qsz) or imgui.is_key_pressed(imgui.Key.escape):
                self.selected_path = None
                hello_imgui.get_runner_params().app_shall_exit = True

            imgui.pop_id()



import os, logging

# this ic import convenciton is from their readme
try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa
    install = None

_debug = bool(int(os.getenv("MBO_DEBUG", "0")))
_level = logging.DEBUG if _debug else logging.INFO

_h = logging.StreamHandler()
_root = logging.getLogger("mbo")
_root.setLevel(_level)
_root.addHandler(_h)
_root.propagate = False


def get(subname: str | None = None) -> logging.Logger:
    name = "mbo" if subname is None else f"mbo.{subname}"
    return logging.getLogger(name)


def enable(*subs):
    for s in subs:
        get(s).disabled = False


def disable(*subs):
    for s in subs:
        get(s).disabled = True


for sub in os.getenv("MBO_ENABLE", "").split(","):
    if sub:
        enable(sub)
for sub in os.getenv("MBO_DISABLE", "").split(","):
    if sub:
        disable(sub)


def get_package_loggers() -> list[str]:
    """Get all loggers that are part of the 'mbo' package."""
    return [
        name
        for name in logging.Logger.manager.loggerDict
        if name.startswith("mbo.")
        and isinstance(logging.Logger.manager.loggerDict[name], logging.Logger)
    ]


def get_all_loggers() -> list[str]:
    """Get all loggers that are currently enabled."""
    return [
        name
        for name, logger in logging.Logger.manager.loggerDict.items()
        if isinstance(logger, logging.Logger) and not logger.disabled
    ]



from __future__ import annotations

import logging
from pathlib import Path
from typing import Sequence, Callable

import numpy as np

from . import log
from .array_types import (
    DemixingResultsArray,
    Suite2pArray,
    H5Array,
    MBOTiffArray,
    TiffArray,
    MboRawArray,
    NpyArray,
)
from .file_io import get_files
from .metadata import is_raw_scanimage, has_mbo_metadata
from .roi import supports_roi

logger = log.get("lazy_array")


SUPPORTED_FTYPES = (
    ".npy",
    ".tif",
    ".tiff",
    ".bin",
    ".h5",
    ".zarr",
)

_ARRAY_TYPE_KWARGS = {
    MboRawArray: {
        "roi",
        "fix_phase",
        "phasecorr_method",
        "border",
        "upsample",
        "max_offset",
    },
    MBOTiffArray: set(),  # accepts no kwargs
    Suite2pArray: set(),  # accepts no kwargs
    H5Array: {"dataset"},
    TiffArray: set(),
    NpyArray: set(),
    DemixingResultsArray: set(),
}


def _filter_kwargs(cls, kwargs):
    allowed = _ARRAY_TYPE_KWARGS.get(cls, set())
    return {k: v for k, v in kwargs.items() if k in allowed}


def imwrite(
    lazy_array,
    outpath: str | Path,
    planes: list | tuple = None,
    roi: int | Sequence[int] | None = None,
    metadata: dict = None,
    overwrite: bool = True,
    ext: str = ".tiff",
    order: list | tuple = None,
    target_chunk_mb: int = 20,
    progress_callback: Callable = None,
    debug: bool = False,
):
    # Logging
    if debug:
        logger.setLevel(logging.INFO)
        logger.info("Debug mode enabled; setting log level to INFO.")
        logger.propagate = True  # send to terminal
    else:
        logger.setLevel(logging.WARNING)
        logger.info("Debug mode disabled; setting log level to WARNING.")
        logger.propagate = False  # don't send to terminal

    # save path
    outpath = Path(outpath)
    if not outpath.parent.is_dir():
        raise ValueError(f"{outpath} is not inside a valid directory.")

    if roi is not None:
        if not supports_roi(lazy_array):
            raise ValueError(
                f"{type(lazy_array)} does not support ROIs, but `roi` was provided."
            )
        lazy_array.roi = roi

    if order is not None:
        if len(order) != len(planes):
            raise ValueError(
                f"The length of the `order` ({len(order)}) does not match the number of planes ({len(planes)})."
            )
        planes = [planes[i] for i in order]

    # Handle metadata
    file_metadata = lazy_array.metadata or {}
    if metadata:
        if not isinstance(metadata, dict):
            raise ValueError(
                f"Provided metadata must be a dictionary, got {type(metadata)} instead."
            )
        file_metadata.update(metadata)

    file_metadata["save_path"] = str(outpath.resolve())
    if hasattr(lazy_array, "metadata"):
        lazy_array.metadata.update(file_metadata)

    if hasattr(lazy_array, "_imwrite"):
        return lazy_array._imwrite(  # noqa
            outpath,
            overwrite=overwrite,
            target_chunk_mb=target_chunk_mb,
            ext=ext,
            progress_callback=progress_callback,
            planes=planes,
            debug=debug,
        )
    else:
        if isinstance(lazy_array, Suite2pArray):
            raise TypeError("Attempting to write a Suite2pArray directly."
                            " Is there an ops.npy file in a directory with a tiff file?"
                            "Please make write these to separate directories.")
        raise TypeError(f"{type(lazy_array)} does not implement an `imwrite()` method.")


def imread(
    inputs: str | Path | Sequence[str | Path],
    **kwargs,  # for the reader
):
    if isinstance(inputs, np.ndarray):
        return inputs
    if isinstance(inputs, MboRawArray):
        return inputs

    if isinstance(inputs, (str, Path)):
        p = Path(inputs)
        if not p.exists():
            raise ValueError(f"Input path does not exist: {p}")
        paths = [Path(f) for f in get_files(p)] if p.is_dir() else [p]
    elif isinstance(inputs, (list, tuple)):
        if isinstance(inputs[0], np.ndarray):
            return inputs
        paths = [Path(p) for p in inputs if isinstance(p, (str, Path))]
    else:
        raise TypeError(f"Unsupported input type: {type(inputs)}")

    if not paths:
        raise ValueError("No input files found.")

    filtered = [p for p in paths if p.suffix.lower() in SUPPORTED_FTYPES]
    if not filtered:
        raise ValueError(f"No supported files in {inputs}")
    paths = filtered

    parent = paths[0].parent if paths else None
    ops_file = parent / "ops.npy" if parent else None

    if ops_file and ops_file.exists():
        md = np.load(str(ops_file), allow_pickle=True).item()
        if (parent / "ops.npy").exists():
            return Suite2pArray(md)
        elif (parent / "reg_tif").is_dir():
            tifs = sorted((parent / "reg_tif").glob("*.tif"))
            if tifs:
                return TiffArray(tifs)
        else:
            raise ValueError(
                f"Unable to open suite2p data from {parent}.\n"
                "Please ensure the directory structure is correct."
            )

    exts = {p.suffix.lower() for p in paths}
    first = paths[0]

    if len(exts) > 1:
        if exts == {".bin", ".npy"}:
            npy_file = first.parent / "ops.npy"
            bin_file = first.parent / "data_raw.bin"
            md = np.load(str(npy_file), allow_pickle=True).item()
            return Suite2pArray(bin_file, md)
        raise ValueError(f"Multiple file types found in input: {exts!r}")

    if first.suffix in [".tif", ".tiff"]:
        if is_raw_scanimage(first):
            return MboRawArray(files=paths, **_filter_kwargs(MboRawArray, kwargs))
        if has_mbo_metadata(first):
            return MBOTiffArray(paths)
        return TiffArray(paths)

    if first.suffix == ".bin":
        npy_file = first.parent / "ops.npy"
        bin_file = first.parent / "data_raw.bin"
        if npy_file.exists():
            md = np.load(str(npy_file), allow_pickle=True).item()
            return Suite2pArray(bin_file, md)
        raise NotImplementedError("BIN files with metadata are not yet supported.")

    if first.suffix == ".h5":
        return H5Array(first)

    if first.suffix == ".npy" and (first.parent / "pmd_demixer.npy").is_file():
        return DemixingResultsArray(first.parent)

    raise TypeError(f"Unsupported file type: {first.suffix}")



class ScanReaderException(Exception):
    """Base ScanReader exception."""

    pass


class ScanImageVersionError(ScanReaderException):
    """Exception for unsupported ScanImage versions."""

    pass


class PathnameError(ScanReaderException):
    """Exception for dealing with paths and pathname patterns (wildcards)."""

    pass


class FieldDimensionMismatch(ScanReaderException):
    """Exception for trying to slice an array with fields of different dimensions."""

    pass
